      function bpower(rk)
c---  Calculates sqrt(Power spectrum) with gaussian smoothing and with
c---  cutoff at Nyquist frequency
c---  Power spectrum consists of primordial scale-free power law and
c---  transfer function
c---  Ref. Eisenstein and Hu, 1998, ApJ, 496,605
      character(3) dark
      common/powsp/rknyq,rkinv,powh,smooth,omehs,dark
      common/cospar/omep,omepb,omeplam,h,ssize
      data in1/0/

c---  Important cosmological scales and epochs
c---  Tcmb = 2.728+-0.0004K (Fixen et al. 1996) adopted
c---  rkeq = scale of particle horizon at equality epoch in unit of [h Mpc^-1]
c---  zd = drag epoch (baryon released from the drag of photons)
c---  sound = sound horizon at drag epoch in [h^-1 Mpc]
c---  rksk = Silk damping scale in unit of [h Mpc^-1]
c     if(in1.eq.0)then
      cmbt27 = 2.728/2.7
      o   = omep*h**2
      ob  = omepb*h**2
      oc  = (omep-omepb)*h**2

      zeq = 2.50e4*o/cmbt27**4
      rkeq = 7.46e-2*omep*h/cmbt27**2
      b1  = 0.313/o**0.419*(1+0.607*o**0.674)
      b2  = 0.238*o**0.223
      zd  = 1291*o**0.251/(1+0.659*o**0.828)*(1+b1*ob**b2)
      Rd  = 31.5*ob/cmbt27**4/(zd/1000.)
      Req = 31.5*ob/cmbt27**4/(zeq/1000.)
      sound=2./3./rkeq*sqrt(6./Req)
     &      *alog((sqrt(1+Rd)+sqrt(Rd+Req))/(1+sqrt(Req)))
      rksk = 1.6*ob**0.52*o**0.73*(1+(10.4*o)**(-0.95))/h

      y = (1+zeq)/(1+zd)
      gy = y*(-6*sqrt(1+y)+(2+3*y)*alog((sqrt(1+y)+1)/(sqrt(1+y)-1)))
      alphab = 2.07*rkeq*sound/(1+Rd)**(3./4.)*gy

      a11  = (46.9*o)**0.670*(1+(32.1*o)**(-0.532))
      a22  = (12.0*o)**0.424*(1+(45.0*o)**(-0.582))
      alphac = a11**(-ob/o)*a22**(-ob/o*ob/o*ob/o)
      b11  = 0.944/(1+(458*o)**(-0.708))
      b22  = 1./(0.395*o)**0.0266
      betac = 1./(1+b11*((oc/o)**b22-1))
      betab = 0.5+ob/o+(3-2*ob/o)*sqrt((17.2*o)**2+1)
      betan = 8.41*o**0.435

      in1 = 1
c     end if

c***  Primordial scale-free power spectrum : power = sqrt(P(k))
c---  rk = 2*pi*(wave number) & rkinv = 1./(2*pi*knyq)
c---  smooth = (smoothing length in pixels /ng) /sqrt(2)
      if(rk.gt.rknyq.or.rk.eq.0.)then
      power=0.
      return
      else
c     power=rk**powh*(1.-(rk*rkinv)**4.)*exp(-(rk*smooth)**2.)
      power=rk**powh
      end if

c***  Baryonic effects included
      q   = rk*omehs*cmbt27**2
      rks = rk/ssize*sound
      f   = 1./(1+(rks/5.4)**4)
      transc = f*tnot(q,1.0,betac)+(1-f)*tnot(q,alphac,betac)

      soundt = sound/(1+(betan/rks)**3)**(1./3.)
      rkstilde = rk/ssize*soundt
      rkrksk = rk/ssize/rksk
      transb = (tnot(q,1.0,1.0)/(1+(rks/5.2)**2)
     & +alphab/(1.+(betab/rks)**3)*exp(-rkrksk**1.4))
     & *sin(rkstilde)/rkstilde
      
      bpower = power*abs(ob*transb+oc*transc)/o

c***  Transfer function (BBKS 1986, p60)
c---  omehs = 1/(omega*h*v[100])
      qq = rk*omehs
c---  Cold dark matter, adiabatic fluctuations with 3 flavors of
c---  relativistic neutrinos plus photons and negligible baryon
c         power=power*alog(1+2.34*qq)/(2.34*qq)*(1+3.89*qq
c    &     +(16.1*qq)**2+(5.46*qq)**3+(6.71*qq)**4)**(-0.25)

      return
      end

      function tnot(q,alphac,betac)
c     data e/2.718281828/
      e = 2.718281828
      c = 14.2/alphac+386./(1+69.9*q**1.08)
      tnot = alog(e+1.8*betac*q)/(alog(e+1.8*betac*q)+c*q**2)
      return
      end

      function bpower1(rk)
c---  Calculates sqrt(Power spectrum) with gaussian smoothing and with
c---  cutoff at Nyquist frequency
c---  Power spectrum consists of primordial scale-free power law and
c---  transfer function
c---  Ref. Eisenstein and Hu, 1998, ApJ, 496,605
      character(3) dark
      common/powsp/rknyq,rkinv,powh,smooth,omehs,dark
      common/cospar/omep,omepb,omeplam,h,ssize
      data in1/0/

c---  Important cosmological scales and epochs
c---  Tcmb = 2.728+-0.0004K (Fixen et al. 1996) adopted
c---  rkeq = scale of particle horizon at equality epoch in unit of [h Mpc^-1]
c---  zd = drag epoch (baryon released from the drag of photons)
c---  sound = sound horizon at drag epoch in [h^-1 Mpc]
c---  rksk = Silk damping scale in unit of [h Mpc^-1]
c     if(in1.eq.0)then
      cmbt27 = 2.728/2.7
      o   = omep*h**2
      ob  = omepb*h**2
      oc  = (omep-omepb)*h**2

      zeq = 2.50e4*o/cmbt27**4
      rkeq = 7.46e-2*omep*h/cmbt27**2
      b1  = 0.313/o**0.419*(1+0.607*o**0.674)
      b2  = 0.238*o**0.223
      zd  = 1291*o**0.251/(1+0.659*o**0.828)*(1+b1*ob**b2)
      Rd  = 31.5*ob/cmbt27**4/(zd/1000.)
      Req = 31.5*ob/cmbt27**4/(zeq/1000.)
      sound=2./3./rkeq*sqrt(6./Req)
     &      *alog((sqrt(1+Rd)+sqrt(Rd+Req))/(1+sqrt(Req)))
      rksk = 1.6*ob**0.52*o**0.73*(1+(10.4*o)**(-0.95))/h

      y = (1+zeq)/(1+zd)
      gy = y*(-6*sqrt(1+y)+(2+3*y)*alog((sqrt(1+y)+1)/(sqrt(1+y)-1)))
      alphab = 2.07*rkeq*sound/(1+Rd)**(3./4.)*gy

      a11  = (46.9*o)**0.670*(1+(32.1*o)**(-0.532))
      a22  = (12.0*o)**0.424*(1+(45.0*o)**(-0.582))
      alphac = a11**(-ob/o)*a22**(-ob/o*ob/o*ob/o)
      b11  = 0.944/(1+(458*o)**(-0.708))
      b22  = 1./(0.395*o)**0.0266
      betac = 1./(1+b11*((oc/o)**b22-1))
      betab = 0.5+ob/o+(3-2*ob/o)*sqrt((17.2*o)**2+1)
      betan = 8.41*o**0.435

      in1 = 1
c     end if

c***  Primordial scale-free power spectrum : power = sqrt(P(k))
c---  rk = 2*pi*(wave number) & rkinv = 1./(2*pi*knyq)
c---  smooth = (smoothing length in pixels /ng) /sqrt(2)
      if(rk.gt.rknyq.or.rk.eq.0.)then
      power=0.
      return
      else
c     power=rk**powh*(1.-(rk*rkinv)**4.)*exp(-(rk*smooth)**2.)
      power=rk**powh
      end if

c***  Baryonic effects included
      q   = rk*omehs*cmbt27**2
      rks = rk/ssize*sound
      f   = 1./(1+(rks/5.4)**4)
      transc = f*tnot(q,1.0,betac)+(1-f)*tnot(q,alphac,betac)

      soundt = sound/(1+(betan/rks)**3)**(1./3.)
      rkstilde = rk/ssize*soundt
      rkrksk = rk/ssize/rksk
      transb = (tnot(q,1.0,1.0)/(1+(rks/5.2)**2)
     &         +alphab/(1.+(betab/rks)**3)*exp(-rkrksk**1.4))
     &         *sin(rkstilde)/rkstilde
      
      bpower1 = power*abs(ob*transb+oc*transc)/o

      return
      end
