#include "defs.h"

extern float yboxup,yboxsize,zboxup,zboxsize;
int myid, numprocs;
MPI_Status receive_status;

INT8 NStackMax;
INT8 NParticles;
INT8 NStack;
INT8 Nmax;
basicparticletype *BasicParticles, *theStack;
basicparticletype particleBlock[P_BLOCK_SIZE];
float WorkAbove, WorkBelow, TotalWork;

SendParticles(hyperLevel,xsplit)
int hyperLevel;  /* Dimension of the hypercube */
float xsplit;
{
	int j;
	int sendprocid, recvprocid;
	int sideflag;
	int count;
	char dummy;
	
	sendprocid = myid;
	recvprocid = sendprocid ^ (hyperLevel >> 1);  
	sideflag = (sendprocid < recvprocid);
		/* holds the essence of the hypercube! */
	count =  0;

/* First stack particles in our list */
	for(j=0; j<NParticles; j++) {
		float x;

		x = BasicParticles[j].z;
		if( (x <= xsplit) == sideflag ) {      /* save the local particles */
/*			
 			theStack[NStack] = BasicParticles[j];
			NStack++;
*/
			BasicParticles[j].u.keepflag = 1;
			if( NStack == NStackMax ) {
                                INT8 nnntmp;
                                nnntmp = NStack+count;
				fprintf(stderr,
					"Proc %d exceeding maximum allowed number %lld in the stack --> %lld\n",myid,NStackMax,nnntmp);
				fprintf(stderr,"P%d Nstack %d, count\n",myid,NStack,count);
				EXIT(0);
			}
		}
		else if ( (x > xsplit) == sideflag ) {  
			particleBlock[count] = BasicParticles[j];
			count++;
			if( count == P_BLOCK_SIZE ) {
				SEND(recvprocid, NP_TAG, &count, sizeof(int));
				SEND(recvprocid, PARTICLE_TAG, particleBlock,
					P_BLOCK_SIZE*sizeof(basicparticletype));
				count = 0;
			}
						/* send other particles across the border */
		}
	}
/* Signal the receiving processor that you are finished sending */
	if( count > 0 ) {
		SEND(recvprocid, NP_TAG, &count, sizeof(int));
		SEND(recvprocid, PARTICLE_TAG, particleBlock,
			count*sizeof(basicparticletype));
	}
	SEND(recvprocid, STOP, &dummy, 0);
}

ReceiveParticles(hyperLevel)
int hyperLevel;
{
	int count;
	int recvprocid, sendprocid;

	recvprocid = myid;
	sendprocid = recvprocid ^ (hyperLevel >> 1);  

/* continue to receive particles until you receive the stop message */
	RECEIVE(sendprocid, ANY_TAG, &count, sizeof(int));
	while( receive_status.MPI_TAG != STOP ) { 
		RECEIVE(sendprocid, ANY_TAG, particleBlock,
						count*sizeof(basicparticletype));
		if( NStack + count >= NStackMax ) {
                        INT8 nnntmp;
                        nnntmp = NStack+count;
			fprintf(stderr,
				"Proc %d exceeding maximum allowed number %lld in the stack --> %lld\n",myid,NStackMax,nnntmp);
			fprintf(stderr,"P%d Nstack %d, count\n",myid,NStack,count);
			EXIT(0);
		}
		memcpy(theStack+NStack,particleBlock,count*sizeof(basicparticletype));
		NStack += count;
		RECEIVE(sendprocid, ANY_TAG, &count, sizeof(int));
	}
}

DomainDecomposition(Np,Bp,domainCorner,domainLength,refcorners,reflengths)
int *Np;
basicparticletype *Bp;
coordStruct *domainCorner;   /* corner the local domain */
coordStruct *domainLength;   /* lengths of the local domain */
coordStruct *refcorners;     /* reference corner of domain hierarchy */
coordStruct *reflengths;     /* reference lengths of domain hierarchy */
{
	int i, j;
	int NProc;
	int nhyper;
	int sendprocid, recvprocid;
	int hyperLevel, axis;
	float xsplit;
	float FindWorkCenter();
	float corner[3], length[3], rmax[3];
	float pcorner[3], plength[3];
	basicparticletype *bp0;

    MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
    MPI_Comm_rank(MPI_COMM_WORLD,&myid);

	NParticles = *Np; BasicParticles = Bp;

	Nmax = NParticles + freespace()/sizeof(basicparticletype) - 32L;
	BasicParticles = (basicparticletype *) 
		resizelast(BasicParticles,Nmax*sizeof(basicparticletype));


	theStack = BasicParticles + NParticles;
	NStackMax = Nmax - NParticles;
	fprintf(stdout,"P%d has %ld particles and Maximum allowed stack is %d\n",myid,NParticles,NStackMax);
	NStack = 0;

	NProc = numprocs;
	SystemMaxExtent(NParticles,BasicParticles,NProc,corner,length); 
			/* find maximum extent of system */
/* initialize reference corners and lengths */
	corner[0] = -0.5; length[0] = 1.0;
	/*
	corner[1] = -0.5; length[1] = 1.0;
	corner[2] = -0.5; length[2] = 1.0;
	*/
	corner[1] = -0.5; length[1] = yboxsize;
	corner[2] = -0.5; length[2] = zboxsize;
	for(i=0; i<3; i++) {
		pcorner[i] = corner[i];
		plength[i] = length[i];
		rmax[i] = corner[i] + length[i];
	}

/* Foreach level in the hypercube begin break down */
	nhyper = 0;
	for(hyperLevel = NProc; hyperLevel >1; hyperLevel = hyperLevel >> 1) {

		for(bp0=BasicParticles; bp0<BasicParticles+(size_t)NParticles; bp0++)
				bp0->u.keepflag = 0;

		axis = 2; /* split on z-axis */
		xsplit = FindWorkCenter(hyperLevel, corner[axis], rmax[axis]);
				/* find the center of work */

		sendprocid = myid;
		recvprocid = sendprocid ^ (hyperLevel >> 1); 
			/* hypercube address protocol */ 

/* condition is necessary to maintain synchronization */
/* send and receive particles from other side of domain */
		if( sendprocid < recvprocid ) {
			SendParticles(hyperLevel,xsplit);
			ReceiveParticles(hyperLevel);
		}
		else {
			ReceiveParticles(hyperLevel);
			SendParticles(hyperLevel,xsplit);
		}
#ifdef DEBUG
		fprintf(stderr,"processor %d  NStack = %d\n", myid, NStack);
#endif
/* Reset the current particles */
		for(i=0, bp0=BasicParticles; bp0<BasicParticles+(size_t)NParticles; bp0++) {
			if( bp0->u.keepflag ) {
				BasicParticles[i] = *bp0;
				i++;
			}
		}
		NParticles = i;
		for(i=NParticles, j=0; i<NParticles+(size_t)NStack; i++, j++) {
				BasicParticles[i] = theStack[j];
		}
		NParticles += NStack;

		NStack = 0; /* rezero */
		theStack = BasicParticles + NParticles;
		NStackMax = Nmax - NParticles;

/* remember the domain hierarchy for this processor */
		if( sendprocid < recvprocid ) {
			plength[axis] = xsplit - pcorner[axis];
		}
		else {
			plength[axis] = plength[axis] + pcorner[axis] - xsplit; 
			pcorner[axis] = xsplit;
		}

/* remember corners and lengths for each level in the hypercube */
		refcorners[nhyper].x = pcorner[0];
		refcorners[nhyper].y = pcorner[1];
		refcorners[nhyper].z = pcorner[2];
		reflengths[nhyper].x = plength[0];
		reflengths[nhyper].y = plength[1];
		reflengths[nhyper].z = plength[2];
		nhyper++;

		SYNCHRONIZE(); /* resynchronize the processors */

	}
	domainCorner->x = pcorner[0];
	domainCorner->y = pcorner[1];
	domainCorner->z = pcorner[2];
	domainLength->x = plength[0];
	domainLength->y = plength[1];
	domainLength->z = plength[2];
#ifdef DEBUG
	if( myid == 0 ) {
	MaximumExtent(NParticles,BasicParticles,corner,rmax);
	length[0] = rmax[0] - corner[0];
	length[1] = rmax[1] - corner[1];
	length[2] = rmax[2] - corner[2];
	fprintf(stderr,"%d %d %g %g %g %g %g %g\nextent %g %g %g %g %g %g\n",
		myid, NParticles,
		pcorner[0], pcorner[1], pcorner[2], plength[0], plength[1], plength[2],
		corner[0], corner[1], corner[2], length[0], length[1], length[2]);
	fflush(stderr);
	}
#endif

	*Np = NParticles;
	BasicParticles = (basicparticletype *) 
		resizelast(BasicParticles,NParticles*sizeof(basicparticletype));

}

/* Find the center of work for each processor */
WeightFrac(x)
float x;
{
	basicparticletype *p;

	WorkAbove = WorkBelow = 0;
	for(p=BasicParticles; p<BasicParticles+(size_t)NParticles; p+=PSKIP) {
		float work;
	
#ifdef VarPM
		if( p->z > x )
			WorkAbove += 1.;
		else
			WorkBelow += 1.;
#else
		work = p->freework;
		if( p->z > x )
			WorkAbove += work;
		else
			WorkBelow += work;
#endif
	}
}

float FindWorkCenter(hyperLevel,xleft,xright)
int hyperLevel;
float xleft, xright;
{
	int iter;
	float xguess, network, eps;
	float RingCombineWork();

	resync(hyperLevel);  
		/* this ensures that all communications are synchronized here */
               /* note it is not a SYNCHRONIZE() */

	xguess = 0.5*(xright + xleft);
	eps = 1;
	network = RingCombineWork(hyperLevel, xguess);
	eps = fabs(network - 0.5);
	iter = 0;
	while( eps > WORKTOL && iter < MAXITER ) {
		if( network > 0.5 ) {
			xright = xguess;
			xguess = 0.5*(xleft + xright);
		}
		else {
			xleft = xguess;
			xguess = 0.5*(xleft + xright);
		}
		network = RingCombineWork(hyperLevel, xguess);
		eps = fabs(network - 0.5);
		iter++;
	}

	resync(hyperLevel);  

	return xguess;
}

float RingCombineWork(hyperLevel,x)
int hyperLevel;
float x;
{
	int i, self, master, src, dest, nsubproc;
	float worksend[2], workrecv[2];
	float totworkabove, totworkbelow;

	nsubproc = hyperLevel;
	self = myid;
	master = (self/nsubproc)*nsubproc;  
		/* Assign master node for the given hyperlevel */
 

	WeightFrac(x); /* Determine the work above and below */

	worksend[0] = WorkAbove; worksend[1] = WorkBelow;
	if( myid == master ) {
		totworkabove = WorkAbove;  totworkbelow = WorkBelow;
		for(i=1; i<nsubproc; i++) {
			RECEIVE(i+master, WORK_TAG, workrecv, 2*sizeof(float));
			totworkabove += workrecv[0];
			totworkbelow += workrecv[1];
		}
		worksend[0] = totworkabove; worksend[1] = totworkbelow;
		for(i=1; i<nsubproc; i++)
			SEND(i+master, WORK_TAG, worksend, 2*sizeof(float));
	}
	else {
		SEND(master, WORK_TAG, worksend, 2*sizeof(float));
		RECEIVE(master, WORK_TAG, workrecv, 2*sizeof(float));
		totworkabove = workrecv[0]; totworkbelow = workrecv[1];
	}

/*
	for(i=1; i<nsubproc; i++) {
		src = self - master - i;
		dest = self - master + i;
		src = (src + nsubproc) % nsubproc;
		dest = (dest + nsubproc) % nsubproc;
		src += master;
		dest += master;

		SENDRECV(src, WORK_TAG, workrecv, 2*sizeof(float), 
			dest, WORK_TAG, worksend, 2*sizeof(float));

		totworkabove += workrecv[0];
		totworkbelow += workrecv[1];
	}
*/
	return(totworkbelow/(totworkabove + totworkbelow));
}

resync(hyperLevel)
int hyperLevel;
{
	int i, partner, done;

    for(i=numprocs>>1; i > 0; i = i >>1 ) {
        partner = myid ^ i; /* hypercube partner */
		if( myid < partner ) {
			SEND(partner, READY_TAG, &done, sizeof(int));
			RECEIVE(partner, READY_TAG, &done, sizeof(int));
		}
		else {
			RECEIVE(partner, READY_TAG, &done, sizeof(int));
			SEND(partner, READY_TAG, &done, sizeof(int));
		}
	}

/*
	if( myid == MASTER_PN ) {
		for(i=1; i<numprocs; i++)
			RECEIVE(i, READY_TAG, &done, sizeof(int));
	}
	else {
			SEND(MASTER_PN, READY_TAG, &done, sizeof(int));
	}
*/
#ifdef DEBUG
	fprintf(stderr,"proc %d hyp %d  Ready\n",myid,hyperLevel);
#endif
	SYNCHRONIZE();         /* get started */
#ifdef DEBUG
	if( myid == 0 )
	fprintf(stderr,"Resync complete at hyperlevel %d\n",hyperLevel);
#endif
}
