#include "defs.h"
#include "extern.h"

float WorkAbove, WorkBelow, TotalWork;

/* Find the center of work for each processor */
WeightFrac(axis,x)
int axis;
float x;
{
	int i, j, n;
	particleStruct *thePart, *p;

	n = NParticles;
	thePart = theParticles;

	WorkAbove = WorkBelow = 0;
	switch( axis ) {

		case XAXIS:
		for(p=theParticles; p<theParticles+n; p+=PSKIP) {
			float work;
	
			work = p->work;
			if( p->r.x > x )
				WorkAbove += work;
			else
				WorkBelow += work;
		}
		break;

		case YAXIS:
		for(p=theParticles; p<theParticles+n; p+=PSKIP) {
			float work;
	
			work = p->work;
			if( p->r.y > x )
				WorkAbove += work;
			else
				WorkBelow += work;
		}
		break;

		case ZAXIS:
		for(p=theParticles; p<theParticles+n; p+=PSKIP) {
			float work;
	
			work = p->work;
			if( p->r.z > x )
				WorkAbove += work;
			else
				WorkBelow += work;
		}
		break;

	}
}

float FindWorkCenter(hyperLevel,xleft,xright,axis)
int hyperLevel;
float xleft, xright;
int axis;
{
	int i, iter;
	int nsubproc, self, master;
	int done;
	float xguess, network, eps;
	float workab[2];
	float RingCombineWork();


	resync(hyperLevel);  
		/* this ensures that all communications are synchronized here */
               /* note it is not a SYNCHRONIZE() */

	xguess = 0.5*(xright + xleft);
	eps = 1;
	network = RingCombineWork(hyperLevel, axis, xguess);
	eps = fabs(network - 0.5);
	iter = 0;
	while( eps > WORKTOL && iter < MAXITER ) {
		if( network > 0.5 ) {
			xright = xguess;
			xguess = 0.5*(xleft + xright);
		}
		else {
			xleft = xguess;
			xguess = 0.5*(xleft + xright);
		}
		network = RingCombineWork(hyperLevel, axis, xguess);
		eps = fabs(network - 0.5);
		iter++;
	}

	resync(hyperLevel);  

	return xguess;
}

float RingCombineWork(hyperLevel,axis,x)
int hyperLevel;
int axis;
float x;
{
	int i, self, master, src, dest, nsubproc;
	float worksend[2], workrecv[2];
	float totworkabove, totworkbelow;

	nsubproc = hyperLevel;
	self = myid;
	master = (self/nsubproc)*nsubproc;  
		/* Assign master node for the given hyperlevel */
 

	WeightFrac(axis, x); /* Determine the work above and below */

	worksend[0] = WorkAbove; worksend[1] = WorkBelow;
	if( myid == master ) {
		totworkabove = WorkAbove;  totworkbelow = WorkBelow;
		for(i=1; i<nsubproc; i++) {
			RECEIVE(i+master, WORK_TAG, workrecv, 2*sizeof(float));
			totworkabove += workrecv[0];
			totworkbelow += workrecv[1];
		}
		worksend[0] = totworkabove; worksend[1] = totworkbelow;
		for(i=1; i<nsubproc; i++)
			SEND(i+master, WORK_TAG, worksend, 2*sizeof(float));
	}
	else {
		SEND(master, WORK_TAG, worksend, 2*sizeof(float));
		RECEIVE(master, WORK_TAG, workrecv, 2*sizeof(float));
		totworkabove = workrecv[0]; totworkbelow = workrecv[1];
	}

/*
	for(i=1; i<nsubproc; i++) {
		src = self - master - i;
		dest = self - master + i;
		src = (src + nsubproc) % nsubproc;
		dest = (dest + nsubproc) % nsubproc;
		src += master;
		dest += master;

		SENDRECV(src, WORK_TAG, workrecv, 2*sizeof(float), 
			dest, WORK_TAG, worksend, 2*sizeof(float));

		totworkabove += workrecv[0];
		totworkbelow += workrecv[1];
	}
*/
	return(totworkbelow/(totworkabove + totworkbelow));
}

resync(hyperLevel)
int hyperLevel;
{
	int i, partner, done;

    for(i=numprocs>>1; i > 0; i = i >>1 ) {
        partner = myid ^ i; /* hypercube partner */
		if( myid < partner ) {
			SEND(partner, READY_TAG, &done, sizeof(int));
			RECEIVE(partner, READY_TAG, &done, sizeof(int));
		}
		else {
			RECEIVE(partner, READY_TAG, &done, sizeof(int));
			SEND(partner, READY_TAG, &done, sizeof(int));
		}
	}

/*
	if( myid == MASTER_PN ) {
		for(i=1; i<numprocs; i++)
			RECEIVE(i, READY_TAG, &done, sizeof(int));
	}
	else {
			SEND(MASTER_PN, READY_TAG, &done, sizeof(int));
	}
*/
#ifdef DEBUG
	fprintf(stderr,"proc %d hyp %d  Ready\n",myid,hyperLevel);
#endif
	SYNCHRONIZE();         /* get started */
#ifdef DEBUG
	if( myid == 0 )
	fprintf(stderr,"Resync complete at hyperlevel %d\n",hyperLevel);
#endif
}
