#include<stdio.h>
#include<math.h>
#include<stdlib.h>
#include<mpi.h>
#include "pmheader.h"
#include "Memory.h"
#define MAX(a,b) ((a)>(b) ? (a):(b))

void G_determinindex(basicparticletype *particles,int np){
	int i;
	INT8 *lnp,tmpnp;
	MPI_Status status;
	int nrank,myrank;

	MPI_Comm_size(MPI_COMM_WORLD,&nrank);
	MPI_Comm_rank(MPI_COMM_WORLD,&myrank);
	lnp = (INT8 *)Malloc(sizeof(INT8)*nrank,PPTR(lnp));
	tmpnp = (INT8)np;
	if(myrank == 0){
		lnp[0] = np;
		for(i=1;i<nrank;i++){
			MPI_Recv(lnp+i,1, MPI_INT8,i,0,MPI_COMM_WORLD,&status);
		}
	}
	else {
		MPI_Send(&tmpnp,1,MPI_INT8,0,0,MPI_COMM_WORLD);
	}
	MPI_Bcast(lnp,nrank,MPI_INT8,0,MPI_COMM_WORLD);
	tmpnp = 0;
	for(i=0;i<myrank;i++)
		tmpnp += lnp[i];

	for(i=0;i<np;i++){
		particles[i].indx = tmpnp;
		tmpnp ++;
	}
	Free(lnp);
}
typedef struct GlacialParticle{
	float x,y,z;
} GlacialParticle;
void InitVel(basicparticletype *particles, int np){
	int i;
	basicparticletype *p;
	p = particles;
	for(i=0;i<np;i++){
		p->vx = p->vy=p->vz = 0.;
		p++;
	}
}
#define MGLA 10000000
int G_np_read(){
	int myid,numprocs,rnp;
	char filename[80];
	int i;
	FILE *fp;
	MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
	sprintf(filename,"Glacial.%.5d",myid);
	for(i=0;i<numprocs;i++){
		if(myid == i){
			fp = fopen(filename,"r");
			fread(&rnp,sizeof(int),1,fp);
		}
	}
	fclose(fp);
	return rnp;
}
int G_read(basicparticletype *particles,float *zstart,float *zheight){
	GlacialParticle *gp;
	int myid,numprocs,rnp;
	basicparticletype *p;
	int i,j,k,count,np;
	char filename[80];
	FILE *fp;
	MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
	sprintf(filename,"Glacial.%.5d",myid);
	gp = (GlacialParticle *) Malloc(sizeof(GlacialParticle)*MGLA,PPTR(gp));
	for(i=0;i<numprocs;i++){
		if(myid == i){
			fprintf(stdout,"P%d start to reading data ...",myid);
			fp = fopen(filename,"r");
			fread(&rnp,sizeof(int),1,fp);
			fread(zstart,sizeof(float),1,fp);
			fread(zheight,sizeof(float),1,fp);
			count = 0;
			p = particles;
			while((np=fread(gp,sizeof(GlacialParticle),MGLA,fp))){
				for(j=0;j<np;j++){
					p->x = gp[j].x;
					p->y = gp[j].y;
					p->z = gp[j].z;
					count ++;
					p++;
				}
			}
			fclose(fp);
			fprintf(stdout,"P%d finishes reading data, %d \n",myid,count);
		}
		fflush(stdout);
		MPI_Barrier(MPI_COMM_WORLD);
	}
	G_determinindex(particles,count);
	Free(gp);
	return count;
}
void G_write(basicparticletype *particles,int np, float zstart,float zheight){
	GlacialParticle *gp;
	int myid,numprocs;
	basicparticletype *p;
	int i,j,k,count;
	char filename[80];
	FILE *wp;
	MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
	sprintf(filename,"Glacial.%.5d",myid);
	gp = (GlacialParticle *) Malloc(sizeof(GlacialParticle)*MGLA,PPTR(gp));
	for(i=0;i<numprocs;i++){
		if(myid == i){
			fprintf(stdout,"P%d start to writing data ...",myid);
			wp = fopen(filename,"w");
			fwrite(&np,sizeof(int),1,wp);
			fwrite(&zstart,sizeof(float),1,wp);
			fwrite(&zheight,sizeof(float),1,wp);
			count = 0;
			p = particles;
			for(j=0;j<np;j++){
				gp[count].x = p->x;
				gp[count].y = p->y;
				gp[count].z = p->z;
				count ++;
				p++;
				if(count == MGLA){
					fwrite(gp,sizeof(GlacialParticle),count,wp);
					count = 0;
				}
			}
			if(count > 0) fwrite(gp,sizeof(GlacialParticle),count,wp);
			count = 0;
			fclose(wp);
			fprintf(stdout,"P%d wrote data ...\n",myid);
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}
	Free(gp);
}

float G_displace(basicparticletype *particles,int np, float pfact){
	int i;
	int myid,numprocs;
	float mdisp,xdisp,ydisp,zdisp,r2disp,tdisp;
	basicparticletype *p;
	MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);

	p = particles;
	mdisp = 0.;
	for(i=0;i<np;i++){
		xdisp = p->vx*pfact;
		ydisp = p->vy*pfact;
		zdisp = p->vz*pfact;
		r2disp = xdisp*xdisp+ydisp*ydisp+zdisp*zdisp;
		mdisp = MAX(mdisp,r2disp);
		p++;
	}
	mdisp = sqrt(mdisp);
	MPI_Reduce(&mdisp,&tdisp,1,MPI_FLOAT,MPI_MAX,0,MPI_COMM_WORLD);
	MPI_Bcast(&tdisp,1,MPI_FLOAT,0,MPI_COMM_WORLD);
	if(myid ==0){
		printf("The Maximum displacement in Glacial initial condition %g\n",tdisp);
	}
	return tdisp;
}
