/*
 * In this code, the MPI_Sendrecv uses MPI_BYTE type instead of somewhat
 * complicated routines.   So if you want to run this code in different
 * OS(for example, little endian and large endian computer system),
 * please use MPI_Aint and MPI_Datatype......   12/22  2000.
 */
#include "defs.h"
#include "extern.h"
#include "pm_extern.h"
#define max(A,B) ( (A) > (B) ? (A) : (B) )
#define min(A,B) ( (A) < (B) ? (A) : (B) )
#define TICKS 100.

void migrate(int *np,int ngz,float zheight,float zstart) {
	INT8 nplocal;
	float zdbound,zubound, ngz2, width;
	int i,j;
	int src,dest,numrecv,numsend;
	MPI_Status status;
	int commsize;
	basicparticletype *psend, *precv, *p;
	INT8 nmax;
	INT8 ntot;

	p = particles;
	fprintf(stderr,"P%d : zheight %g zstart %g\n",myrank,zheight,zstart);
#ifdef DEBUG
if( myrank == 0 ) {
	fprintf(stderr,"zheight %g zstart %g\n",zheight,zstart);
	fprintf(stderr,"velocities %g %g %g\n",p->vx, p->vy, p->vz);
}
#endif
TIMER_START(33);
	nplocal = *np;
	nmax = nplocal + freespace()/sizeof(basicparticletype) -32LL;
        fprintf(stdout,"nmax in migration %d\n",nmax);
      
	particles = (basicparticletype *) 
			resizelast(particles,nmax*sizeof(basicparticletype));
#ifdef DEBUG
	fprintf(stderr,"P%d nplocal %lld nmax %lld\n",myrank,nplocal,nmax);
#endif

	/*
	 * These two lines are to reduce the call on Operating system for
	 * more memory.  So reduce or free current massive memory usage
	 * into free list.
	*/

	zdbound = zstart;
	zubound = zstart+zheight;
#ifdef DEBUG
fprintf(stderr,"P%d %g < z < %g\n",myrank,zdbound,zubound);
#endif
	width = zubound - zdbound;
	if( ngz == 1 )
		ngz2 = 0;
	else
		ngz2 = 0.5*ngz;

/* Send up */
	commsize = 1;
	while( commsize > 0 ) {
		psend = particles + nplocal;
		i = 0; j = 0;
		if( myrank == 0 ) {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z >= zubound && p->z < ngz2 ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		if( myrank == nrank-1 ) {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z < ngz2 ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		if( myrank > 0 && myrank < nrank-1)  {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z >= zubound ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		numsend = i;
		numrecv = numsend;
		nplocal = j;
		precv = psend + numsend;

		dest = (myrank + 1 + nrank) % nrank;
		src =  (myrank - 1 + nrank) % nrank;

	    MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,NUMRECV_TAG,
             src,NUMRECV_TAG,MPI_COMM_WORLD,&status);
/*
	    MPI_Sendrecv(psend,numsend*sizeof(basicparticletype),MPI_BYTE,dest,
					PSEND_TAG,
					precv,numrecv*sizeof(basicparticletype),MPI_BYTE,src,
					PSEND_TAG,MPI_COMM_WORLD,&status);
		MPI_Allreduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
		if( myrank == 0 ) {
			fprintf(stderr,"Sending %d up\n",commsize);
		}
*/
        MPI_Allreduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
        if( myrank == 0 ) {
            fprintf(stderr,"Sending %d up  ...",commsize);
        }
        for(i=0;i<nrank;i++){
            MPI_Barrier(MPI_COMM_WORLD);
            if(myrank == i)
                MPI_Send(psend,numsend*sizeof(basicparticletype),MPI_BYTE,
                        dest,PSEND_TAG,MPI_COMM_WORLD);
            else if(myrank == (i+1)%nrank)
                MPI_Recv(precv,numrecv*sizeof(basicparticletype),MPI_BYTE,src,
                        PSEND_TAG,MPI_COMM_WORLD,&status);
        }
	fprintf(stdout,"after communication in migrate %d<-->%d\n",numrecv,numsend);
        if( myrank == 0 ) {
            fprintf(stderr,"Sent %d up\n",commsize);
        }


		i = 0;
		for(p=particles+nplocal; p<particles+nplocal+numrecv; p++) {
			*p = precv[i];
			i++;
		}
		nplocal += numrecv;
	}

/* Sending down */
	commsize = 1;
	while( commsize > 0 ) {
		psend = particles + nplocal;
		i = 0; j = 0;
		if( myrank == 0 ) {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z > ngz2 ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		if( myrank == nrank - 1) {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z < zdbound && p->z > ngz2 ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		if( myrank > 0 && myrank < nrank-1)  {
			for(p=particles; p<particles+nplocal; p++) {
				if( p->z < zdbound ) {
					psend[i] = *p;
					i++;
				}
				else {
					particles[j] = *p;
					j++;
				}
			}
		}
		numsend = i;
		numrecv = numsend;
		nplocal = j;
		precv = psend + numsend;

		dest = (myrank - 1 + nrank) % nrank;
		src =  (myrank + 1 + nrank) % nrank;
	    MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,NUMRECV_TAG,
             src,NUMRECV_TAG,MPI_COMM_WORLD,&status);
/*
	    MPI_Sendrecv(psend,numsend*sizeof(basicparticletype),MPI_BYTE,dest,
					PSEND_TAG,
					precv,numrecv*sizeof(basicparticletype),MPI_BYTE,src,
					PSEND_TAG,MPI_COMM_WORLD,&status);
		MPI_Allreduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
		if( myrank == 0 ) {
			fprintf(stderr,"Sending %d down\n",commsize);
		}
*/
        MPI_Allreduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,MPI_COMM_WORLD);
        if( myrank == 0 ) {
            fprintf(stderr,"Sending %d down   ",commsize);
        }
        for(i=0;i<nrank;i++){
            if(myrank == i) 
                MPI_Send(psend,numsend*sizeof(basicparticletype),MPI_BYTE,dest,
                        PSEND_TAG,MPI_COMM_WORLD);
            else if(myrank == (i-1+nrank)%nrank)
                MPI_Recv(precv,numrecv*sizeof(basicparticletype),MPI_BYTE,src,
                        PSEND_TAG,MPI_COMM_WORLD,&status);
            MPI_Barrier(MPI_COMM_WORLD);
        }
        if( myrank == 0 ) {
            fprintf(stderr,"Sent %d down\n",commsize);
        }


		i = 0;
		for(p=particles+nplocal; p<particles+nplocal+numrecv; p++) {
			*p = precv[i];
			i++;
		}
		nplocal += numrecv;
	}

/* Check */

#ifdef DEBUG
	for(p=particles; p<particles+nplocal; p++) {
		if( p->z < zdbound || p->z >= zubound) {
			fprintf(stderr,"p->z %g out of bounds in P%d\n",
				p->z,myrank);
		}
	}
#endif

	MPI_Allreduce(&nplocal,&ntot,1,MPI_INT8,MPI_SUM,MPI_COMM_WORLD);
	fprintf(stderr,"P%d ntot %lld nplocal %lld\n",myrank,ntot,nplocal);
	particles = (basicparticletype *) 
			resizelast(particles,nplocal*sizeof(basicparticletype));
	*np = nplocal;

	SYNCHRONIZE();
	TIMER_STOP(33);
	if( myrank == 0 )
		fprintf(stderr,"CPU(migrate)=%g\n",ELAPSED_TIME(33));
}
