/* My tree correction algorithm */
/* Modified from size_t to long integer due to the problem
 * occuring when comparison is made using signed and unsigned integers
 * 10/12/2005
 * */
#include<stdio.h>
#include<stdlib.h>
#include<stddef.h>
#include<math.h>
#include "mpi.h"
#include "Memory.h"
#include "pmheader.h"
#include "kjhtree.h"
#include "force_spline.h"
#define min(a,b) ((a)<(b) ? (a):(b))
#define max(a,b) ((a)>(b) ? (a):(b))
#define CellWidth 4L /* The cellwidth is preferred to be same as rsphere */
#define rsphere 4.
#define nstride 1
float rspheresq;
typedef struct {
	float x,y,z;
} ComGhost;
long npmaxcell,tnpmaxcell,ncalcell;
long Ncorrect;
static initflag=0;
long mx,my,mz;
basicparticletype **BasicCell;
long *NBasicCell;
long npos,ngrv,ntmp;
particle *pos,*tpos;
TPtlStruct *grv,*tgrv;
TStruct *TREECELL;
ComGhost *cghost;
void Ghost(long,float,float,float,int,int,int);
void i_force_spline(long,float);
int dest,src;
int myid,nid;
float theta2;
void ParticleCellLinkedList(basicparticletype *particles, long np,
		float zstart){
	long i,ix,iy,iz,mpixel;
	basicparticletype *p,*tp;

	for(i=0;i<mx*my*mz;i++){
		BasicCell[i] = NULL;
		NBasicCell[i] = 0;
	}
	tp = particles;
	for(i=0;i<np;i++){
		ix = (long)(tp->x/CellWidth);
		iy = (long)(tp->y/CellWidth);
		iz = (long)((tp->z-zstart)/CellWidth);
		mpixel = ix+mx*(iy+my*iz);
		p = BasicCell[mpixel];
		BasicCell[mpixel] = tp;
		tp->u.nextinbin = p;
		NBasicCell[mpixel] ++;
		tp++;
	}
}
Box CheckBoundingBox(){
	float xmin,xmax,ymin,ymax,zmin,zmax;
	float width;
	long i;
	Box box;
	xmin = ymin = zmin = 1.e25;
	xmax = ymax = zmax = -1.e25;
	for(i=0;i<ngrv;i++){
		xmin = min(xmin,grv[i].r[0]);
		ymin = min(ymin,grv[i].r[1]);
		zmin = min(zmin,grv[i].r[2]);
		xmax = max(xmax,grv[i].r[0]);
		ymax = max(ymax,grv[i].r[1]);
		zmax = max(zmax,grv[i].r[2]);
	}
	width = xmax-xmin;
	width = max(width,ymax-ymin);
	width = max(width,zmax-zmin);
	if(width !=0){
		box.x = xmin-width*1.e-6;
		box.y = ymin-width*1.e-6;
		box.z = zmin-width*1.e-6;
		box.width = width*(1.+2.e-6);
	}
	else {
		box.x = xmin-1.e-5;
		box.y = ymin-1.e-5;
		box.z = zmin-1.e-5;
		box.width = 2.e-5;
	}
	return box;
}
void WringGrvInfo(long ncell){
	basicparticletype *p;
	long ntmp;
	p = BasicCell[ncell];
	ntmp = 0;
	while(p){
		grv[ntmp].r[0] = p->x;
		grv[ntmp].r[1] = p->y;
		grv[ntmp].r[2] = p->z;
		p = p->u.nextinbin;
		ntmp ++;
	}
}
long PosCountParticle(long ix0,long iy0,long iz0,long imx,
		long ipx,long imy,long ipy,long imz,long ipz) {
	long ncount,mcell;
	long ix,iy,iz;
	long ix2,iy2,iz2;
	ncount = 0;
	for(iz=iz0+imz;iz<iz0+ipz;iz++){
		if(iz < 0 || iz >= mz) continue;
		else iz2 = iz;
		for(iy=iy0+imy;iy<iy0+ipy;iy++){
			iy2 = (iy+my)%my;
			for(ix=ix0+imx;ix<ix0+ipx;ix++){
				ix2 = (ix+mx)%mx;
				mcell = ix2+mx*(iy2+my*iz2);
				ncount += NBasicCell[mcell];
			}
		}
	}
	return ncount;
}
/* Dump2Grv is designed to copy particle data in grv format considering isolated boundaries */
long GrvCountParticle(long ix0,long iy0,long iz0,long ixm,
		long ixp,long iym,long iyp,long izm,long izp) {
	long ix,iy,iz;
	long ixi,iyi,izi,ixf,iyf,izf;
	long mcell,ncount;
	/* for isolated boundary conditions */
	ixi = max(ix0+ixm,0);
	iyi = max(iy0+iym,0);
	izi = max(iz0+izm,0);
	ixf = min(ix0+ixp,mx);
	iyf = min(iy0+iyp,my);
	izf = min(iz0+izp,mz);

	ncount = 0;
	for(iz=izi;iz<izf;iz++){
		for(iy=iyi;iy<iyf;iy++){
			for(ix=ixi;ix<ixf;ix++){
				mcell = ix+mx*(iy+my*iz);
				ncount += NBasicCell[mcell];
			}
		}
	}
	return ncount;

}
void Dump2Grv(long ix0,long iy0,long iz0,long ixm,long ixp,
		long iym,long iyp,long izm,long izp) {
	long ix,iy,iz,mcell;
	long ixi,iyi,izi,ixf,iyf,izf;
	long ntmp;
	basicparticletype *p;
	tgrv = grv;
	ntmp = 0;
	/* for isolated boundary conditions */
	ixi = max(ix0+ixm,0);
	iyi = max(iy0+iym,0);
	izi = max(iz0+izm,0);
	ixf = min(ix0+ixp,mx);
	iyf = min(iy0+iyp,my);
	izf = min(iz0+izp,mz);

	for(iz=izi;iz<izf;iz++){
		for(iy=iyi;iy<iyf;iy++){
			for(ix=ixi;ix<ixf;ix++){
				mcell = ix+mx*(iy+my*iz);
				p=BasicCell[mcell];
				while(p){
					grv[ntmp].type = TYPE_PTL;
					grv[ntmp].r[0] = p->x;
					grv[ntmp].r[1] = p->y;
					grv[ntmp].r[2] = p->z;
					p = p->u.nextinbin;
					ntmp++;
				}
			}
		}
	}
}

void Dump2PeriodicPosExceptZ(long ix0,long iy0,long iz0,long ixm,
		long ixp, long iym, long iyp, long izm, long izp,int nx,int ny){
	long ix,iy,iz;
	long ix2,iy2,iz2;
	long mcell;
	float lpx,lpy;
	basicparticletype *p;
	tpos = pos;
	for(iz=iz0+izm;iz<iz0+izp;iz++){
		if(iz < 0 || iz >= mz) continue;
		else iz2 = iz;
		for(iy=iy0+iym;iy<iy0+iyp;iy++){
			iy2 = (iy+my)%my;
			if(iy > iy2) lpy = ny;
			else if(iy < iy2) lpy = -ny;
			else lpy = 0.;
			for(ix=ix0+ixm;ix<ix0+ixp;ix++){
				ix2 = (ix+mx)%mx;
				mcell = ix2+mx*(iy2+my*iz2);
				p=BasicCell[mcell];
				if(ix > ix2) lpx = nx;
				else if(ix < ix2) lpx = -nx;
				else lpx = 0.;
				while(p){
					tpos->x = p->x + lpx;
					tpos->y = p->y + lpy;
					tpos->z = p->z;
					tpos->bp = p;
					tpos ++;
					p = p->u.nextinbin;
				}
			}
		}
	}
}
/*
void determine_mpi_long(){
	if(sizeof(long)==sizeof(unsigned int)) {
		MPI_long = MPI_UNSIGNED;
	}
	else if(sizeof(long) == sizeof(unsigned long)){
		MPI_long = MPI_UNSIGNED_LONG;
	}
}
*/
long MaxAllowableCell(){
	long nfree,ncell,needed,additional,nscale;

	nfree = freespace();
	/* needed for gravitating source particles */
	needed = (sizeof(TStruct)+sizeof(TPtlStruct)+sizeof(particle)*27)*npmaxcell;

	return ncell;
}
/* v(t+dt/2) = fact1 * v(t-dt/2) + fact2 + a(t-dt/2) 
 * r(t+dt) = r(t-dt) + v(t+dt/2)*pfact
 */
void UpdateVelocityUsingForceByParticles(long np, float vfact2){
	long i;
	for(i=0;i<np;i++){
		(pos[i].bp)->vx += vfact2*pos[i].ax;
		(pos[i].bp)->vy += vfact2*pos[i].ay;
		(pos[i].bp)->vz += vfact2*pos[i].az;
	}
}
extern float diff[NSPLINE][3],slope[NSPLINE][3],ran2nran;
void DirectSummation(){
	float tmpx,tmpy,tmpz,fplmf;
	float px,py,pz,dist2,ptlmass,dist;
	long i,j,ntmp;
	for(i=0;i<npos;i++)
		pos[i].ax = pos[i].ay = pos[i].az = 0.;
	
	for(i=0;i<npos;i++){
		px = pos[i].x; py = pos[i].y; pz = pos[i].z;
		for(j=0;j<ngrv;j++){
			tmpx = px - grv[j].r[0];
			tmpy = py - grv[j].r[1];
			tmpz = pz - grv[j].r[2];
			dist2 = tmpx*tmpx + tmpy*tmpy + tmpz*tmpz;
			if(dist2 <= rspheresq){
				dist = sqrt(dist2);
				ntmp = dist/ran2nran;
				fplmf = diff[ntmp][0] + slope[ntmp][0] * (dist-ntmp*ran2nran);  
				pos[i].ax += tmpx* fplmf;
				pos[i].ay += tmpy* fplmf;
				pos[i].az += tmpz* fplmf;
			}
		}
	}
}
void PartialUpdateAccelUsingForceByParticles(){
	long i;
	basicparticletype *bp;
	float xmin,ymin,zmin;
	float xmax,ymax,zmax;
	/*
	xmin=ymin=zmin=1.e23;
	xmax=ymax=zmax=-1.e23;
	*/
	for(i=0;i<npos;i++){
		bp = pos[i].bp;
		bp->ax += pos[i].ax;
		bp->ay += pos[i].ay;
		bp->az += pos[i].az;
		/*
		xmax = max(xmax,pos[i].ax);
		ymax = max(ymax,pos[i].ay);
		zmax = max(zmax,pos[i].az);
		xmin = min(xmin,pos[i].ax);
		ymin = min(ymin,pos[i].ay);
		zmin = min(zmin,pos[i].az);
		*/
	}
	/*
	printf("P%d has x=%g %g y=%g %g z=%g %g\n",myid,xmin,xmax,ymin,ymax,zmin,zmax);
	*/
}
void AllUpdateVelocityAndPosition(basicparticletype *particles,long np,
		int nx, int ny, int nz,
		float pfact, float vfact1, float vfact2){
	long i;
	basicparticletype *tp;
	tp = particles;
	for(i=0;i<np;i++){
		tp->vx = vfact1* tp->vx + vfact2 * tp->ax;
		tp->vy = vfact1* tp->vy + vfact2 * tp->ay;
		tp->vz = vfact1* tp->vz + vfact2 * tp->az;
		tp->x += pfact * tp->vx;
		tp->y += pfact * tp->vy;
		tp->z += pfact * tp->vz;
		tp++;
	}
	tp = particles;
	for(i=0;i<np;i++){
		tp->x = fmod(tp->x+nx,nx);
		tp->y = fmod(tp->y+ny,ny);
		tp->z = fmod(tp->z+nz,nz);
	}
}

int treecorrection(basicparticletype *particles,long np, int nx, int ny, 
		int nz,
		float zstart, float zheight, float vfact1, float vfact2,
		float pfact,float theta){
	long i,j,k;
	long ix0,iy0,iz0;
	long ix2,iy2,iz2;
	long jxp,jyp,jzp;
	long ncell,mcell,mp;
	basicparticletype *p;
	Box box;
	rspheresq = rsphere*rsphere;

	/*
	determine_mpi_long();
	*/
	MPI_Comm_size(MPI_COMM_WORLD,&nid);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);

	theta2 = theta*theta*0.25;



	mx = (long)ceil((double)nx/(double)CellWidth);
	my = (long)ceil((double)ny/(double)CellWidth);
	mz = (long)ceil((double)zheight/(double)CellWidth);

	if(initflag==0) {
		if(nx != ny && nx!=nz) {
			if(myid==0) {
				fprintf(stderr,"Error in TreeCorrection\n");
				fprintf(stderr,"since nx != ny, we don't know how to set force scaling \n");
			}
			exit(99);
		}
		printf("P%d Now initializing force array\n",myid);
		i_force_spline(nx,rsphere);
		initflag = 1;
	}
	printf("P%d has mx=%d my=%d mz=%d memory for basicell = %ld\n",myid,mx,my,mz,
			sizeof(basicparticletype *)*mx*my*mz);

	BasicCell = (basicparticletype **) 
		Malloc(sizeof(basicparticletype *)*mx*my*mz,PPTR(BasicCell));
	NBasicCell = (long *) Malloc(sizeof(long)*mx*my*mz,PPTR(NBasicCell));
	ParticleCellLinkedList(particles,np,zstart);


	for(iz0=0;iz0<mz;iz0+=nstride){
		jzp = min(iz0+nstride,mz)-iz0+1;
		for(iy0=0;iy0<my;iy0+=nstride){
			jyp = min(iy0+nstride,my)-iy0+1;
			printf("P%d now passing iy0=%d iz0=%d\n",myid,iy0,iz0);
			for(ix0=0;ix0<mx;ix0+=nstride){
				jxp = min(ix0+nstride,mx)-ix0+1;
				/* Making tree buffer */
				ngrv = GrvCountParticle(ix0,iy0,iz0,0,nstride,0,nstride,0,nstride);
				if(ngrv > 0){
					npos = PosCountParticle(ix0,iy0,iz0,-1,jxp,-1,jyp,-1,jzp);
					pos = (particle *)Malloc(sizeof(particle)*npos,PPTR(pos));
					Dump2PeriodicPosExceptZ(ix0,iy0,iz0,-1,jxp,-1,jyp,-1,jzp,nx,ny);
					grv = (TPtlStruct *) Malloc(sizeof(TPtlStruct)*ngrv,PPTR(grv));
					TREECELL = (TStruct *)Malloc(sizeof(TStruct)*ngrv,PPTR(TREECELL));
					Dump2Grv(ix0,iy0,iz0,0,nstride,0,nstride,0,nstride);
					if(ngrv >= DIRECTSUM){
						for(i=0;i<ngrv;i++){
							grv[i].sibling = &grv[i+1];
							grv[i].mass = 1;
						}
						grv[ngrv-1].sibling = NULL;
						box = CheckBoundingBox();
	
						Make_Tree(TREECELL,grv,ngrv,box);
						for(i=0;i<npos;i++){
							treeforce(pos+i,theta2,TREECELL,grv,rspheresq);
						}
					}
					else {
						DirectSummation();
					}
					Free(TREECELL);
					Free(grv);
					PartialUpdateAccelUsingForceByParticles();
					Free(pos);
				}
			}
		}
	}
	/*
	{
		basicparticletype *p;
		FILE *fp;
		if(myid==0){
			fp=fopen("tmp.dat","w");
			for(i=0;i<mx*my*mz;i++){
				p=BasicCell[i];
				while(p){
				if(p->ax != 0. || p->ay != 0. ||p->az != 0. )
					fprintf(fp,"%g %g %g %g %g %g\n",p->x,p->y,p->z,p->ax,p->ay,
							p->az);
				p = p->u.nextinbin;
				}
			}
			fclose(fp);
		}
		exit(9);
	}
	*/
	MPI_Barrier(MPI_COMM_WORLD);
	printf("P%d now entering ghost routine 1\n",myid);
	Ghost(1,zstart,zheight,vfact2,nx,ny,nz);
	MPI_Barrier(MPI_COMM_WORLD);
	printf("P%d now entering ghost routine -1\n",myid);
	Ghost(-1,zstart,zheight,vfact2,nx,ny,nz);
	Free(NBasicCell);
	Free(BasicCell);
	{
		/*
		basicparticletype *bp;
		float xmin,ymin,zmin;
		float xmax,ymax,zmax;
		int indx;
		xmin=ymin=zmin=1.e23;
		xmax=ymax=zmax=-1.e23;
		bp = particles;
		for(i=0;i<np;i++){
			xmax = max(xmax,bp[i].ax);
			ymax = max(ymax,bp[i].ay);
			zmax = max(zmax,bp[i].az);
			xmin = min(xmin,bp[i].ax);
			ymin = min(ymin,bp[i].ay);
			zmin = min(zmin,bp[i].az);
		}
		printf("MINMAX P%d has x=%g %g y=%g %g z=%g %g\n",myid,xmin,xmax,ymin,ymax,zmin,zmax);
		for(i=0;i<nid;i++){
			FILE *fp;
			MPI_Barrier(MPI_COMM_WORLD);
			if(myid==i){
				if(myid==0)fp=fopen("tmp.dat","w");
				else fp=fopen("tmp.dat","a");
				for(j=0;j<np;j++){
					if(bp[j].az < -6.80e-07 || bp[j].az > 6.80e-07){
						fprintf(fp,"%g %g %g %g %g %g\n",bp[j].x,bp[j].y,bp[j].z,
								bp[j].ax,bp[j].ay,bp[j].az);
					}
				}
				fclose(fp);
			}

		}
		*/
	}
}

/* if iflag == 1 --> upper sending 
 * else --> down sending
 * */
void Ghost(long iflag,float zstart,float zheight,float vfact2,
		int nx, int ny, int nz){
	long i,j,k;
	long nrghost,nsghost,mcell;
	long ix,iy,iz;
	long ix0,iy0,izg0;
	long jxp,jyp,jzm,jzp;
	long ix1,iy1,izg1;
	long iz0pos;
	long ncount;
	long mstride;
	int sndtag,recvtag;
	MPI_Status status;
	float z_start,z_stop;
	float lpz;
	basicparticletype *p;
	Box box;
	ComGhost *sghost,*rghost;


	mstride = 4;
	sndtag = recvtag = 1;
	if(iflag == 1) {
		dest = (myid+1+nid)%nid;
		src = (myid-1+nid)%nid;
		izg1 = mz-1;/* plus slice (being included) for gravitation data */
		if(fmod(zheight,rsphere)==0.L) izg0 = izg1;
		else izg0 = izg1-1;
		if(mz==1) izg0 = 0;
		z_start = zstart+zheight-rsphere;
		z_stop = zstart+zheight;
		iz0pos = 0;
		jzm = 0; /* minus slice for position data */
		jzp = 1; /* plus slice for position data */
	}
	else {
		dest = (myid-1+nid)%nid;
		src = (myid+1+nid)%nid;
		izg1 = izg0 = 0;
		z_start = zstart;
		z_stop = zstart+rsphere;
		iz0pos = mz-1;
		if(fmod(zheight,rsphere)==0.L){
			jzm = 0;
			jzp = 1;
		}
		else{
			jzm = -1;
			jzp = 1;
		}
	}
	printf("P%d has izg0=%d izg1=%d\n",myid,izg0,izg1);
	for(iy0=0;iy0<my;iy0+=mstride){
		if(myid==0) printf("P%d has ix0=%d iy0=%d \n",myid,ix0,iy0);
		iy1 = min(my,iy0+mstride);
		jyp = iy1-iy0+1;
		for(ix0=0;ix0<mx;ix0+=mstride){
			ix1 = min(mx,ix0+mstride);
			jxp = ix1-ix0+1;
			nsghost = 0;
			for(iz=izg0;iz<=izg1;iz++)
				for(iy=iy0;iy<iy1;iy++)
					for(ix=ix0;ix<ix1;ix++){
						mcell = ix+mx*(iy+my*iz);
						nsghost += NBasicCell[mcell];
					}
			if(nsghost > 0){
				sghost = (ComGhost*)Malloc(sizeof(ComGhost)*nsghost,PPTR(sghost));
				if(iflag ==1 && myid==nid-1) lpz = -nz;
				else if(iflag == -1 && myid== 0) lpz = nz;
				else lpz = 0;
				ncount = 0;
				for(iz=izg0;iz<=izg1;iz++)
					for(iy=iy0;iy<iy1;iy++)
						for(ix=ix0;ix<ix1;ix++){
							mcell = ix + mx*(iy+my*iz);
							p = BasicCell[mcell];
							while(p){
								if(p->z >=z_start && p->z < z_stop){
									sghost[ncount].x = p->x;
									sghost[ncount].y = p->y;
									sghost[ncount].z = p->z + lpz;
									ncount++;
								}
								p= p->u.nextinbin;
							}
						}
				nsghost = ncount;
				sghost = (ComGhost*)Realloc(sghost,sizeof(ComGhost)*nsghost);
			}
			else nsghost = 0;
			MPI_Sendrecv(&nsghost,1,MPI_LONG,dest,sndtag,
					&nrghost,1,MPI_LONG,src,recvtag,MPI_COMM_WORLD,
					&status);
			if(nrghost >0) rghost = (ComGhost*)Malloc(sizeof(ComGhost)*nrghost,PPTR(rghost));
			MPI_Sendrecv(sghost,sizeof(ComGhost)*nsghost,MPI_BYTE,dest,sndtag,
					rghost,sizeof(ComGhost)*nrghost,MPI_BYTE,src,recvtag,
					MPI_COMM_WORLD,&status);
			if(nsghost > 0) Free(sghost);
			ngrv = nrghost;
			if(ngrv > 0){
				grv = (TPtlStruct *) Malloc(sizeof(TPtlStruct)*ngrv,PPTR(grv));
				TREECELL = (TStruct *)Malloc(sizeof(TStruct)*ngrv,PPTR(TREECELL));
				for(i=0;i<ngrv;i++){
					grv[i].type = TYPE_PTL;
					grv[i].r[0] = rghost[i].x;
					grv[i].r[1] = rghost[i].y;
					grv[i].r[2] = rghost[i].z;
					grv[i].sibling = &grv[i+1];
					grv[i].mass = 1;
				}
				grv[ngrv-1].sibling = NULL;
				Free(rghost);
				npos = PosCountParticle(ix0,iy0,iz0pos,-1,jxp,-1,jyp,jzm,jzp);
				if(npos > 0){
					pos = (particle *)Malloc(sizeof(particle)*npos,PPTR(pos));
					Dump2PeriodicPosExceptZ(ix0,iy0,iz0pos,-1,jxp,-1,jyp,jzm,jzp,nx,ny);
		
					box = CheckBoundingBox();
					Make_Tree(TREECELL,grv,ngrv,box);
					for(i=0;i<npos;i++){
						treeforce(pos+i,theta2,TREECELL,grv,rspheresq);
					}
					/*
					UpdateVelocityUsingForceByParticles(pos,npos,vfact2);
					*/
					PartialUpdateAccelUsingForceByParticles();
					Free(pos);
				}
				Free(TREECELL);
				Free(grv);
			}
			else Free(rghost);

		}
	}
}


