#include "defs.h"
#include "extern.h"
#define BYTE_BLOCK_SIZE 1048576
/*
#define INFINITY -1
*/

/* number of cells to zero */ 
#define ZeroCells(n,p0) \
{ \
	int n0; \
	char *c, *p; \
 \
	p = (char *) p0; \
 \
	n0 = (int) (n); \
	for(c=p; c<p+n0; c++) \
		*c = 0x0; \
}

cellStruct *MakeNewCell()  /* make a single new cell */
{
	cellStruct *theCell;
	int lastcell;

	lastcell = NCells + 1;
	if ( lastcell > NCellMax ) {
		fprintf(stderr,
	    "Proc %d Exceeded maximum allowed number of cells - NCells = %d\n", 
			myid,lastcell);
		EXIT(1);
	}

	theCell = &theCells[NCells];
	ZeroCells(sizeof(cellStruct),theCell);
	theCell->type    = CELL_TYPE;
	theCell->epssq   = 0.0;

	NCells++;

	return theCell;
}

int ParticleAllocate(np)
int np;
{
	INT8 nmax;
	
    theParticles = (particleStruct *) Malloc(INFINITY,PPTR(theParticles));
	nmax = ptrsize(theParticles)/sizeof(particleStruct) - 32;
	if( nmax < np ) {
		fprintf(stderr,"way too many particles - ParticleAllocate aborting\n");
		
		EXIT(0);
	}
    if( theParticles == NULL ) {
        fprintf(stderr,"Cannot allocate memory for particles in processor %d\n",myid);
		fprintf(stderr,"Current available memory size is %ld",
				CheckAvailableMemory()/sizeof(char));fflush(stderr);
		fprintf(stderr,"and INFINITY is %ld",INFINITY);fflush(stderr);
        EXIT(1);
    }
	return nmax;
}

/* if you run out of space for particles double the local capacity */
ExpandParticleMemory()
{
    fprintf(stderr,"P%d cannot reallocate memory for more particles\n", myid);
       EXIT(1);
}

InitCellMemory(n)
int n;
{

	NCells = 0;
    theTreeNodes = (cellStruct **)
        Malloc(2L*(INT8)n*sizeof(cellStruct *),PPTR(theTreeNodes));
    if( theTreeNodes == NULL ) {
        fprintf(stderr,
        "Cannot allocate memory for tree nodes %d in processor %d\n", n,myid);
   
        EXIT(1);
    }
    theGroupNodes = (particleStruct **)
        Malloc((INT8)n*sizeof(particleStruct *),PPTR(theGroupNodes));
    if( theGroupNodes == NULL ) {
        fprintf(stderr,
        "Cannot allocate memory for group nodes in processor %d\n", myid);
   
        EXIT(1);
    }

	NCellMax = n;
	if( NCellMax < 100 ) NCellMax = 1000;
	theCells = (cellStruct *) Malloc((INT8)NCellMax*sizeof(cellStruct),PPTR(theCells));
    if( theCells == NULL ) {
        fprintf(stderr,
        "Cannot allocate memory for the cells in processor %d\n", myid);
		
        EXIT(1);
	}
}

FreeTreeMemory()
{
	freelast(theCells);
	freelast(theGroupNodes);
	freelast(theTreeNodes);
	freelast(theParticles);
}
