#include "defs.h"
#include "extern.h"

#define MEGABYTE 1048576
#ifndef NMEG
#define NMEG     40
#endif
#define NPTRMAX  32
#define INFINITY -1

static int startmemory;
int nptr, psize[NPTRMAX], offset[NPTRMAX];
char *Memory;
char *ptr[NPTRMAX];

char *mymalloc(size)
int size;
{
	int newoffset;
	int nbytesleft;
	char *newptr;

	if( startmemory ) {
		if( size >= NMEG*MEGABYTE) {
			fprintf(stderr,"Can't allocate %d Mbytes\n",size);
			return NULL;
		}
		ptr[0] = Memory;
		if( size == INFINITY ) {
			size = NMEG*MEGABYTE;
		}
		else {
			size += (size % 8); /* enforce multiples of 8 bytes */
		}
		psize[0] = size;
		offset[0] = 0;
		nptr   = 1;
		startmemory = 0;
		return ptr[0];
	}
	newoffset = offset[nptr-1] + psize[nptr-1];
	nbytesleft = NMEG*MEGABYTE - newoffset;
	if( size == INFINITY ) {
		size = nbytesleft;
	}
	else {
		size += (size % 8);
	}
/*
	if( myid == 0 ) fprintf(stderr,"nbytes left is %d\n",nbytesleft);
*/
	if( nbytesleft < size ) {
		fprintf(stderr,"mymalloc out of memory- exiting\n");
		return NULL;
	}
	newptr = ptr[nptr-1] + psize[nptr-1];
	psize[nptr] = size;
	offset[nptr] = newoffset;
	ptr[nptr] = newptr;
	nptr++;
	return newptr;
}

dumpptr()
{
	int i;
	if( myid == 0 ) {
	for(i=0; i<nptr; i++)
		fprintf(stderr,"%d %d %d\n",i,ptr[i], ptr[0]+offset[i]);
	}
}

char *resizelast(p,newsize)
char *p;
int newsize;
{
	int freespace();

	if( p != ptr[nptr-1] ) {
		fprintf(stderr,"this is not the last pointer - can't resize it\n");
		EXIT(0);
	}

	if( newsize == INFINITY ) {
		newsize = psize[nptr-1] + freespace();
		newsize += (newsize % 8);
		psize[nptr-1] = newsize;
		return p;
	}
	else {
		newsize += (newsize % 8);
		psize[nptr-1] = newsize;
		return p;
	}
}
	

freelast(p)
char *p;
{
	if( p != ptr[nptr-1] ) {
		fprintf(stderr,"this is not the last pointer - can't free it\n");
		EXIT(0);
	}
	else {
		ptr[nptr-1] = NULL;
		nptr--;
	}
}

int ptrsize(p)
char *p;
{
	int i;

	for(i=0; i<nptr; i++) if( p == ptr[i] ) break;

	return psize[i];
}

int freespace()
{
	int i, bytes;

	bytes = 0;
	for(i=0; i<nptr; i++) {
		bytes += psize[i];
	}

	return NMEG*MEGABYTE - bytes;
}

int InitializeBigMemory()
{
	Memory = (char *) malloc(NMEG*MEGABYTE);
	if( Memory == NULL ) {
		fprintf(stderr,"Can't allocate %d Mbytes\n",NMEG);
		return 0;
	}
	startmemory = 1;
	return 1;
}
