#include<stdio.h>
#include<math.h>
#include<mpi.h>
#include<stdlib.h>
#include<sys/times.h>
#define max(A,B) ( (A) > (B) ? (A) : (B) )
#define min(A,B) ( (A) < (B) ? (A) : (B) )
void fresize(int);
extern float *mass,*x,*y,*z,*vx,*vy,*vz,*work;
extern int *indx;
extern int densize;
extern float *den;
#define TICKS 100.
extern float ctimes();
void migrate(int *np,int myrank,int nrank,int local_nz,int local_z_start,
		int ngz)
{
  	float tmprv[6];
	int ntmp,zdbound,zubound,i,j,unp,zp;
	int nget,nput,src,dest,numrecv,ierror,numsend,nend,npold;
	MPI_Status status;
	int commsize,zup,zdown;
	float *array[6];
	int *intwork;
	float *fltwork;

	/*
	 * These two lines are to reduce the call on Operating system for
	 * more memory.  So reduce or free current massive memory usage
	 * into free list.
	*/
	/*
  	work = (float *)dlrealloc(work,densize*sizeof(float));
	den = work;
	*/

	float sstep;

    if(myrank == 0) sstep = ctimes();

	intwork = (int *) work;
	fltwork = (float *) work;

	array[0] = x;
	array[1] = y;
	array[2] = z;
	array[3] = vx;
	array[4] = vy;
	array[5] = vz;
	
	zdbound = local_z_start;
	zubound = local_z_start+local_nz;
	do {
		unp = *np;
		/*
		 *  These lines are for the case
		 *  **********-----, where - is for migration.
		 */
		i = unp;
		do {
			i = i - 1;
    		zp = (int) *(z+i);
	    	zup = fabs(zubound-zp);
	    	zup = min(zup,ngz-zup);
	    	zdown = fabs(zdbound-zp);
	    	zdown = min(zdown,ngz-zdown);
		} while( (zp >= zubound || zp < zdbound) && zup <= zdown );
		unp = i + 1;
		/*
		 */
		i = 0;
        while(i < unp){
			zp = (int) *(z+i);
 	      	zup = fabs(zubound - zp);
        	zup = min(zup,ngz-zup);
			zdown = fabs(zdbound-zp);
			zdown = min(zdown,ngz-zdown);
			if((zp >= zubound || zp < zdbound) && zup <= zdown){
				for(j=0;j<6;j++) tmprv[j] = *(array[j]+i);
				ntmp = *(indx+i);
				do {
					unp = unp - 1;
					zp = (int) *(z+unp);
					zup = fabs(zubound-zp);
					zup = min(zup,ngz-zup);
					zdown = fabs(zdbound-zp);
					zdown = min(zdown,ngz-zdown);
				} while((zp>=zubound || zp<zdbound) && zup<=zdown && unp>i);
				/* exchange two elements */
				for(j=0;j<6;j++) *(array[j]+i) = *(array[j]+unp);
				*(indx+i) = *(indx+unp);
				for(j=0;j<6;j++) *(array[j]+unp) = tmprv[j];
				*(indx+unp) = ntmp;
			}
			i++;
		}
		src = (myrank-1+nrank) % nrank;
		dest = (myrank+1+nrank) % nrank;
  		numsend = *np-unp;
		numrecv = numsend;
		MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,1,
				src,1,MPI_COMM_WORLD,&status);
		/*
		 * resize memory(x,y,z,vx,vy,vz,indx);
		 */
		/*
		 * total # of particles to send;
		 */
   		MPI_Reduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
   		MPI_Bcast(&commsize,1,MPI_INT,0,MPI_COMM_WORLD);
		if(commsize != 0){
		    nend = max(numrecv+unp,*np);
      		fresize(nend);
        	array[0] = x;
        	array[1] = y;
        	array[2] = z;
        	array[3] = vx;
        	array[4] = vy;
        	array[5] = vz;
	     	if(myrank == 0) printf("total upper sending .... %d \n",commsize);
			for(i=0;i<6;i++){
				MPI_Sendrecv(array[i]+unp,numsend,MPI_FLOAT,dest,i,work,
						numrecv,MPI_FLOAT,src,i,MPI_COMM_WORLD,&status);
				for(j=0;j<numrecv;j++) *(array[i]+unp+j)=*(work+j);
			}
			MPI_Sendrecv((indx+unp),numsend,MPI_INT,dest,i,intwork,
					numrecv,MPI_INT,src,i,MPI_COMM_WORLD,&status);
			for(j=0;j<numrecv;j++) *(indx+unp+j)=*(intwork+j);
		    *np = *np-numsend+numrecv;
      		fresize(*np);
	        array[0] = x;
        	array[1] = y;
        	array[2] = z;
        	array[3] = vx;
        	array[4] = vy;
        	array[5] = vz;
		}
	} while(commsize != 0);

	do {
		unp = *np;
		/*
		 *  These lines are for the case
		 *  **********-----, where - is for migration.
		 */
		i = unp;
		do {
			i = i - 1;
    		zp = (int) *(z+i);
		} while(zp >= zubound || zp < zdbound);
		unp = i + 1;
		/*
		 */
		i = 0;
        while( i < unp){
			zp = (int) *(z+i);
			if((zp >= zubound || zp < zdbound)){
				for(j=0;j<6;j++) tmprv[j] = *(array[j]+i);
				ntmp = *(indx+i);
				do {
					unp = unp - 1;
					zp = (int) *(z+unp);
				} while((zp >= zubound || zp < zdbound)&& unp > i) ;
				/* change two elements */
				for(j=0;j<6;j++) *(array[j]+i) = *(array[j]+unp);
				*(indx+i) = *(indx+unp);
				for(j=0;j<6;j++) *(array[j]+unp) = tmprv[j];
				*(indx+unp) = ntmp;
			}
			i++;
		}
		src = (myrank+1+nrank)%nrank;
		dest = (myrank-1+nrank)%nrank;
  		numsend = *np-unp;
		numrecv = numsend;
		MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,1,
		     src,1,MPI_COMM_WORLD,&status);
		/*
		 * resize memory(x,y,z,vx,vy,vz,indx);
		 */
		/*
		 * total # of particles to send;
		 */
   		MPI_Reduce(&numsend,&commsize,1,MPI_INT,MPI_SUM, 0,MPI_COMM_WORLD);
   		MPI_Bcast(&commsize,1,MPI_INT,0,MPI_COMM_WORLD);
	    fltwork = (float *) work;
		if(commsize != 0){
		    nend = max(numrecv+unp,*np);
      		fresize(nend);
        	array[0] = x;
        	array[1] = y;
        	array[2] = z;
        	array[3] = vx;
        	array[4] = vy;
        	array[5] = vz;
	     	if(myrank == 0) printf("total down sending .... %d \n",commsize);
			for(i=0;i<6;i++){
				MPI_Sendrecv(array[i]+unp,numsend,MPI_FLOAT,dest,i,work,
						numrecv,MPI_FLOAT,src,i,MPI_COMM_WORLD,&status);
				for(j=0;j<numrecv;j++) *(array[i]+unp+j)=*(work+j);
			}
			MPI_Sendrecv((indx+unp),numsend,MPI_INT,dest,0,intwork,
					numrecv,MPI_INT,src,0,MPI_COMM_WORLD,&status);
			for(j=0;j<numrecv;j++) *(indx+unp+j)=*(intwork+j);
			/*
			 * resize memory
			 */
		    *np = *np-numsend+numrecv;
      		fresize(*np);
		}
	} while(commsize != 0);
	/*
	printf("%d %d %f %f \n",myrank,*np,*(z+*np-2),*(z+*np-1));
	fflush(stdout);
	MPI_Finalize();
	exit(0);
	*/
    if(myrank == 0) printf("  CPU(migrate) %G \n",ctimes()-sstep);
}
#include<unistd.h>
void fresize(int np)
{
    x = (float *)realloc(x,np*sizeof(float));
    y = (float *)realloc(y,np*sizeof(float));
    z = (float *)realloc(z,np*sizeof(float));
    vx = (float *)realloc(vx,np*sizeof(float));
    vy = (float *)realloc(vy,np*sizeof(float));
    vz = (float *)realloc(vz,np*sizeof(float));
    indx = (int *)realloc(indx,np*sizeof(int));
}
)realloc(z,np*sizeof(float));
    vx = (float *)realloc(vx,np*sizeof(float));
    vy = (float *)realloc(vy,np*sizeof(float));
    vz = (float *)realloc(vz,np*sizeof(float));
    indx = (int *)realloc(indx,np*sizeof(int));
}
