/*
 * In this code, the MPI_Sendrecv uses MPI_BYTE type instead of somewhat
 * complicated routines.   So if you want to run this code in different
 * OS(for example, little endian and large endian computer system),
 * please use MPI_Aint and MPI_Datatype......   12/22  2000.
 */
#include<stdio.h>
#include<math.h>
#include<mpi.h>
#include<stdlib.h>
#include<sys/times.h>
#include "pmheader.h"
#define max(A,B) ( (A) > (B) ? (A) : (B) )
#define min(A,B) ( (A) < (B) ? (A) : (B) )
void fresize(int);
/*
extern float *mass,*x,*y,*z,*vx,*vy,*vz;
extern int *indx;
*/
extern float *work;
extern basicparticletype *particles;
extern int densize;
extern float *den;
#define TICKS 100.
extern float ctimes();
int blocklens[]={1,1,1,1,1,1,1,1,1,1,1};
MPI_Datatype old_types[]={MPI_FLOAT,MPI_FLOAT,MPI_FLOAT,
	        MPI_FLOAT,MPI_FLOAT,MPI_FLOAT,MPI_FLOAT,MPI_FLOAT,MPI_FLOAT,
	        MPI_INT,MPI_UB};
MPI_Datatype MPI_BASIC_PARTICLE;
MPI_Aint indices[11];

void migrate(int *np,int myrank,int nrank,int local_nz,int local_z_start,
		int ngz)
{
  	basicparticletype tmpparticle;
	int ntmp,zdbound,zubound,i,j,unp,zp;
	int nget,nput,src,dest,numrecv,ierror,numsend,nend,npold;
	MPI_Status status;
	int commsize,zup,zdown;
	int *intwork;
	float *fltwork;
	/*
	float *tmpz;
	*/
	basicparticletype *work_particles;

    indices[0] = 0;
    indices[1] = sizeof(float);
    indices[2] = sizeof(float)+indices[1];
    indices[3] = sizeof(float)+indices[2];
    indices[4] = sizeof(float)+indices[3];
    indices[5] = sizeof(float)+indices[4];
    indices[6] = sizeof(float)+indices[5];
    indices[7] = sizeof(float)+indices[6];
    indices[8] = sizeof(float)+indices[7];
    indices[9] = sizeof(int)+indices[8];
    indices[10] = sizeof(basicparticletype);

	MPI_Type_struct(11,blocklens,indices,old_types,&MPI_BASIC_PARTICLE);

	work_particles = (basicparticletype *)work;


	/*
	tmpz = (float *)malloc(*np*sizeof(float));
	for(i=0;i<*np;i++) tmpz[i] = particles[i].z;
	*/

	/*
	 * These two lines are to reduce the call on Operating system for
	 * more memory.  So reduce or free current massive memory usage
	 * into free list.
	*/
	/*
  	work = (float *)dlrealloc(work,densize*sizeof(float));
	den = work;
	*/

	float sstep;

    if(myrank == 0) sstep = ctimes();

	intwork = (int *) work;
	fltwork = (float *) work;

	zdbound = local_z_start;
	zubound = local_z_start+local_nz;
	do {
		unp = *np; 
		/*
		 *  These lines are for the case
		 *  **********-----, where - is for migration.
		 */
		i = unp; 

		do {
			i = i - 1; 
			zp = (int) particles[i].z;
			/*
			zp = (int) tmpz[i];
			*/
	    	zup = abs(zubound-zp);
	    	zup = min(zup,ngz-zup);
	    	zdown = abs(zdbound-zp);
	    	zdown = min(zdown,ngz-zdown);
		} while( (zp >= zubound || zp < zdbound) && zup <= zdown );
		unp = i + 1;	
		/*
		 */
		i = 0;	
        while(i < unp){
			zp = (int) particles[i].z;
			/*
			zp = (int) tmpz[i];
			*/
 	      	zup = abs(zubound - zp);
        	zup = min(zup,ngz-zup);
			zdown = abs(zdbound-zp);
			zdown = min(zdown,ngz-zdown);
			if((zp >= zubound || zp < zdbound) && zup <= zdown){
				/*
				copyparticle(&tmpparticle,&particles[i]);
				*/
				tmpparticle = particles[i];
				do {
					unp = unp - 1;  
					zp = (int) particles[unp].z;
					zup = abs(zubound-zp);
					zup = min(zup,ngz-zup);
					zdown = abs(zdbound-zp);
					zdown = min(zdown,ngz-zdown);
				} while((zp>=zubound || zp<zdbound) && zup<=zdown && unp>i);
				/* exchange two elements */
				/*
				copyparticle(&particles[i],&particles[unp]);
				*/
				particles[i] = particles[unp];
				/*
				tmpz[i] = tmpz[unp];
				*/
				/*
				copyparticle(&particles[unp],&tmpparticle);
				*/
				particles[unp] = tmpparticle;
				/*
				tmpz[unp] = tmpparticle.z;
				*/
			}
			i++;
		}
		src = (myrank-1+nrank) % nrank;
		dest = (myrank+1+nrank) % nrank;
  		numsend = *np-unp;
		numrecv = numsend;
		MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,1,
				src,1,MPI_COMM_WORLD,&status);
		/*
		 * resize memory(x,y,z,vx,vy,vz,indx);
		 */
		/*
		 * total # of particles to send;
		 */
   		MPI_Reduce(&numsend,&commsize,1,MPI_INT,MPI_SUM,0,MPI_COMM_WORLD);
   		MPI_Bcast(&commsize,1,MPI_INT,0,MPI_COMM_WORLD);
		if(commsize != 0){
		    nend = max(numrecv+unp,*np);
			particles = (basicparticletype *)
				realloc(particles,nend*sizeof(basicparticletype));
	     	if(myrank == 0) printf("total upper sending .... %d \n",commsize);
			/*
			MPI_Sendrecv(&particles[unp],numsend*sizeof(basicparticletype),
					MPI_BYTE,dest,0,work_particles,
					numrecv*sizeof(basicparticletype),
					MPI_BYTE,src,0,MPI_COMM_WORLD,&status);
					*/
			MPI_Sendrecv(&particles[unp],numsend,MPI_BASIC_PARTICLE,
					dest,1,work_particles,numrecv,MPI_BASIC_PARTICLE,
					src,1,MPI_COMM_WORLD,&status);
			/*
			copyparticles(&particles[unp],&work_particles[0],numrecv);
			*/
			memcpy(&particles[unp],&work_particles[0],
					numrecv*sizeof(basicparticletype));
			/*
			tmpz = realloc(tmpz,nend*sizeof(float));
			for(j=0;j<numrecv;j++)
				tmpz[unp+j] = work_particles[j].z;
			*/
		    *np = *np-numsend+numrecv;
			particles = (basicparticletype *) realloc(
					particles,*np*sizeof(basicparticletype));
		}
	} while(commsize != 0);

	do {
		unp = *np;
		/*
		 *  These lines are for the case
		 *  **********-----, where - is for migration.
		 */
		i = unp;
		do {
			i = i - 1;
			zp = (int) particles[i].z;
			/*
			zp = (int) tmpz[i];
			*/
		} while(zp >= zubound || zp < zdbound);
		unp = i + 1;
		/*
		 */
		i = 0;
        while( i < unp){
			zp = (int) particles[i].z;
			/*
			zp = (int) tmpz[i];
			*/
			if((zp >= zubound || zp < zdbound)){
				/*
				copyparticle(&tmpparticle,&particles[i]);
				*/
				tmpparticle = particles[i];
				do {
					unp = unp - 1;
					zp = (int) particles[unp].z;
				} while((zp >= zubound || zp < zdbound)&& unp > i) ;
				/* change two elements */
				/*
				copyparticle(&particles[i],&particles[unp]);
				*/
				particles[i] = particles[unp];
				/*
				tmpz[i] = tmpz[unp];
				*/
				/*
				copyparticle(&particles[unp],&tmpparticle);
				*/
				particles[unp] = tmpparticle;
				/*
				tmpz[unp] = tmpparticle.z;
				*/
			}
			i++;
		}
		src = (myrank+1+nrank)%nrank;
		dest = (myrank-1+nrank)%nrank;
  		numsend = *np-unp;
		numrecv = numsend;
		MPI_Sendrecv_replace(&numrecv,1,MPI_INT,dest,1,
		     src,1,MPI_COMM_WORLD,&status);
		/*
		 * resize memory(x,y,z,vx,vy,vz,indx);
		 */
		/*
		 * total # of particles to send;
		 */
   		MPI_Reduce(&numsend,&commsize,1,MPI_INT,MPI_SUM, 0,MPI_COMM_WORLD);
   		MPI_Bcast(&commsize,1,MPI_INT,0,MPI_COMM_WORLD);
	    fltwork = (float *) work;
		if(commsize != 0){
		    nend = max(numrecv+unp,*np);
			particles = (basicparticletype *) realloc(particles,
					nend*sizeof(basicparticletype));
	     	if(myrank == 0) printf("total down sending .... %d \n",commsize);
			/*
			MPI_Sendrecv(&particles[unp],numsend*sizeof(basicparticletype),
					MPI_BYTE,dest,0,work_particles,
					numrecv*sizeof(basicparticletype),
					MPI_BYTE,src,0,MPI_COMM_WORLD,&status);
					*/
			MPI_Sendrecv(&particles[unp],numsend,MPI_BASIC_PARTICLE,
					dest,1,work_particles,numrecv,MPI_BASIC_PARTICLE,
					src,1,MPI_COMM_WORLD,&status);
			/*
			copyparticles(&particles[unp],&work_particles[0],numrecv);
			*/
			memcpy(&particles[unp],&work_particles[0],
					numrecv*sizeof(basicparticletype));
			/*
			tmpz = realloc(tmpz,nend*sizeof(float));
			for(j=0;j<numrecv;j++)
				tmpz[unp+j] = work_particles[j].z;
				*/

			/*
			 * resize memory
			 */
		    *np = *np-numsend+numrecv;
			particles = (basicparticletype *) realloc(particles,
					*np*sizeof(basicparticletype));
			/*
      		fresize(*np);
			*/
		}
	} while(commsize != 0);
    if(myrank == 0) printf("  CPU(migrate) %G \n",ctimes()-sstep);
	/*
	free(tmpz);
	*/
}
#include<unistd.h>
void fresize(int np)
{
	particles = (basicparticletype *) realloc(particles,
			np*sizeof(basicparticletype));
	/*
    x = (float *)realloc(x,np*sizeof(float));
    y = (float *)realloc(y,np*sizeof(float));
    z = (float *)realloc(z,np*sizeof(float));
    vx = (float *)realloc(vx,np*sizeof(float));
    vy = (float *)realloc(vy,np*sizeof(float));
    vz = (float *)realloc(vz,np*sizeof(float));
    indx = (int *)realloc(indx,np*sizeof(int));
	*/
}
