#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>
#include "pmheader.h"
#define CCD_TAG 20
#define MASTER_PN 0
#define XY 0
#define XZ 1
#define YZ 2
#define SYNCHRONIZE() MPI_Barrier(MPI_COMM_WORLD);

typedef struct {
	float r[3];
} vec;

typedef struct {
	char prefix[20];
	int movieskip;
	int nx, ny;
	int istart, np;
	int orient;
	float xc, yc;
	float boxsize;
} movie;

int myid, numprocs;
static int nframe;
static movie frame[20];
static float *ccd0, *ccd1;

int initframe()
{
	int i;
	int nxmax, nymax;
	movie *m;
	FILE *mvinit;

	MPI_Comm_size(MPI_COMM_WORLD,&numprocs);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);

	if( myid == 0 ) {
		if( !(mvinit = fopen("movie.init","r")) ) {
			nframe = 9999;
			MPI_Bcast(&nframe, 1, MPI_INT, MASTER_PN, MPI_COMM_WORLD);
			return 0;
		}
#ifdef VERBOSE
		fprintf(stderr,"opening movie.init ....\n");
#endif

		m = &frame[0]; nframe = 0;
		while( fscanf(mvinit,"%s\n",m->prefix) != EOF ) {
			fscanf(mvinit,"%d\n",&(m->movieskip));
			fscanf(mvinit,"%d %d\n",&(m->nx),&(m->ny));
			fscanf(mvinit,"%d %d\n",&(m->istart),&(m->np));
			fscanf(mvinit,"%d\n",&(m->orient));
			fscanf(mvinit,"%f %f\n",&(m->xc),&(m->yc));
			fscanf(mvinit,"%f\n",&(m->boxsize));
			nframe++; m++;
		}
		MPI_Bcast(&nframe, 1, MPI_INT, MASTER_PN, MPI_COMM_WORLD);
		MPI_Bcast(frame, nframe*sizeof(movie), MPI_BYTE, 
			MASTER_PN, MPI_COMM_WORLD);
	}
	else {
		MPI_Bcast(&nframe, 1, MPI_INT, MASTER_PN, MPI_COMM_WORLD);
		if( nframe == 9999 ) return 0;
		MPI_Bcast(frame, nframe*sizeof(movie), MPI_BYTE,
			MASTER_PN, MPI_COMM_WORLD);
	}
#ifdef VERBOSE
	fprintf(stderr,"P%d\n",myid);
		for(i=0; i<nframe; i++) {
			m = &frame[i];
			fprintf(stderr,"***  MOVIE %d  ***\n",i);
			fprintf(stderr,"PREFIX: %s\n",m->prefix);
			fprintf(stderr,"MSKIP : %d\n",m->movieskip);
			fprintf(stderr,"PIXELS: %d %d\n",m->nx,m->ny);
			fprintf(stderr,"IS  NP: %d %d\n",m->istart,m->np);
			fprintf(stderr,"ORIENT: %d\n",m->orient);
			fprintf(stderr,"XC, YC: %g %g\n",m->xc,m->yc);
			fprintf(stderr,"BOXSIZ: %g\n",m->boxsize);
		}
#endif 
			
	nxmax = frame[0].nx;
	nymax = frame[0].ny;
	for(i=0; i<nframe; i++) {
		m = &frame[i];
		if( m->nx > nxmax ) nxmax = m->nx;
		if( m->ny > nymax ) nymax = m->ny;
	}
	ccd0 = (float *) calloc(nxmax*nymax,sizeof(float));
	ccd1 = (float *) calloc(nxmax*nymax,sizeof(float));

	SYNCHRONIZE();
	return 1;
}

mkframe(stepnum1,t1,nobj1,particles)
int *stepnum1;
float *t1;
int *nobj1;
basicparticletype  *particles;
{
	int nobj, stepnum;
	float t;
	int i, nx, ny, count;
	char movieframe[80];
	FILE *framefile;
	void mkgs();

	stepnum = *stepnum1;
	nobj = *nobj1;
	t = *t1;
	for(i=0; i<nframe; i++) {
		if( stepnum % frame[i].movieskip == 0 ) {
#ifdef VERBOSE
	fprintf(stderr,"making first image\n");
#endif
			mkgs(nobj,particles,&frame[i]);
			count = frame[i].nx*frame[i].ny;
SYNCHRONIZE();
			(void) MPI_Reduce(ccd0, ccd1, count, MPI_FLOAT, MPI_SUM, 
				MASTER_PN, MPI_COMM_WORLD);
SYNCHRONIZE();
		}
		if( myid == 0 ) {
			sprintf(movieframe,"%s.%.4d",frame[i].prefix,stepnum);
			framefile = fopen(movieframe,"w");
			nx = frame[i].nx; ny = frame[i].ny;
			fwrite(&nx,sizeof(int),1,framefile);
			fwrite(&ny,sizeof(int),1,framefile);
			fwrite(ccd1,nx*ny*sizeof(float),1,framefile);
			fclose(framefile);
		}
	}
	SYNCHRONIZE();
}

void mkgs(nobj,particles,m)
int nobj;
basicparticletype *particles;
movie *m;
{
	int i, ix, iy,id;
	int nx, ny, orient, istart, np;
	float xc, yc, boxsize;
	basicparticletype *p;
	void zero();

	nx = m->nx; ny = m->ny;
	istart = m->istart; np = m->np;
	boxsize = m->boxsize; xc = m->xc; yc = m->yc;
	orient = m->orient;


	zero(ccd0,nx*ny*sizeof(float));
/* Bin the data into the CCD frame */
	for(i=0; i<nobj; i++) {
		p = &particles[i];
/*
		id = r[i].idnum;
		if( id < istart || id >= istart + np) continue;
*/
		switch( orient ) {
			case XY:
			ix = (int) ((p->x - xc + 0.5*boxsize)/boxsize*nx);
			iy = (int) ((p->y - yc + 0.5*(boxsize*ny)/nx)/boxsize*nx);
			iy = ny - iy - 1;
			break;

			case XZ:
			ix = (int) ((p->x - xc + 0.5*boxsize)/boxsize*nx);
			iy = (int) ((p->z - yc + 0.5*(boxsize*ny)/nx)/boxsize*nx);
			iy = ny - iy - 1;
			break;

			case YZ:
			ix = (int) ((p->y - xc + 0.5*boxsize)/boxsize*nx);
			iy = (int) ((p->z - yc + 0.5*(boxsize*ny)/nx)/boxsize*nx);
			iy = ny - iy - 1;
			break;
		}
		if( ix < 0 || ix > nx ) continue;
		if( iy < 0 || iy > ny ) continue;

		(*(ccd0 + ix + nx*iy)) += 1.0;
	}
#ifdef VERBOSE
fprintf(stderr,"P%d exiting mkgs\n",myid);
#endif
}

void zero(c,size)
char *c;
int size;
{
	char *p;
	for(p=c; p<c+size; p++) *p = 0;
}
