/* My GOTPM code using the tree corrections developed by Juhan Kim
 * =======================================================================
 * It is in an interim code for the 4096^3 particle simulations at PSC.
 ***************************************************************************
 * In mytreecorrection, there is no preprocessor for the sake of simplicity.
 * parpmseed_force.F is not yet implemented into namu.c
 ***************************************************************************
 * Comparison to the original code has been checked with a success. 
 * (Jan. 17th 2006)
 * Time evolution has been checked ? (Jan. 18th 2006)
 * FoF group test has been passed ? (Jan. 19th 2006)
 * FoF mass function has been checked when z=0 (Mar. 12th 2006)
 * Simulated matter power spectrum has been checked when z=0 & z=47 
 * (Mar. 13th 2006)
 */
#include<stdio.h>
#include<math.h>
#include<mpi.h>
#include<stddef.h>
#include <stdlib.h>
#include<sys/times.h>
#include "pmheader.h"
#include "common_simparam.h"
#include "pm_common.h"
#include "Memory.h"
#include "varpm.h"
#include <srfftw_mpi.h>

#define MAXINHOMO 1. /* in % */
#define max(A,B) ( (A) > (B) ? (A) : (B) )
#define min(A,B) ( (A) < (B) ? (A) : (B) )
#define pow2(A) ( (A)*(A))
#define pow3(A) ( (A)*(A)*(A))
void migrate(size_t *,int ,float,float);
int flagpsmeasure(float , float , float ,int );
int flagsyncwholeden(float , float , float );
float cputime0[100],cputime1[100];
float gettime(),elapsedtime;

void pm(int, int, int, int, int, int);
int nxpos,nsizepmparticletype;
int nlzfloor,nlzceil,local_nz_varpm;
float lzfloor,lzceil;
int now,nowflag;
int myid,nid;
float omep,omeplam,size,astep,zi,omei,redshift,hubble;
float zheight,zstart,kenold;
float poten;
float *den,*fftwmesh;
int iseed,local_z_start,local_ny_after_transpose;
int local_y_start_after_transpose,total_local_size;
int stepcount;
int nforcepos;
float pmas;

float pfact,fact1,fact2;
float nextastep;
int nstep=0;
void jwrite(int );
int syncflag;
typedef struct HALF{
	int first,second;
}HALF;
HALF halfstep;

#define DomainDecomp {\
	int npmin,npmax,npav;\
	long nptot,lnp;\
	float ratio1,ratio2;\
	lnp = np;\
	MPI_Reduce(&lnp,&nptot,1,MPI_LONG,MPI_SUM,0,MPI_COMM_WORLD);\
	if(myid==0) npav = nptot/nid;\
	MPI_Bcast(&npav,1,MPI_INT,0,MPI_COMM_WORLD);\
	MPI_Reduce(&np,&npmin,1,MPI_INT,MPI_MIN,0,MPI_COMM_WORLD);\
	MPI_Reduce(&np,&npmax,1,MPI_INT,MPI_MAX,0,MPI_COMM_WORLD);\
	MPI_Bcast(&npmin,1,MPI_INT,0,MPI_COMM_WORLD);\
	MPI_Bcast(&npmax,1,MPI_INT,0,MPI_COMM_WORLD);\
	ratio1 = (float)(npav-npmin)/(float)npav*100.;\
	ratio2 = (float)(npmax-npav)/(float)npav*100.;\
	if(myid==0)printf("The inhomogeneities are %g %g\n",ratio1,ratio2);\
	if(ratio1 > MAXINHOMO || ratio2 > MAXINHOMO) {\
		long tnp,lnp;\
		float zmin,zmax;\
		int mydomaindecomposition(pmparticletype *, int ,long ,\
				        float ,float , float *, float *,int );\
		TIMER_START(39)\
		lnp = np;\
		MPI_Reduce(&lnp,&tnp,1,MPI_LONG,MPI_SUM,0,MPI_COMM_WORLD);\
		MPI_Bcast(&tnp,1,MPI_LONG,0,MPI_COMM_WORLD);\
		zmin = zstart; zmax = zstart+zheight;\
		np= mydomaindecomposition(pmparticles, np,tnp,zmin,zmax, &zmin, &zmax,nz);\
		zstart = zmin; zheight = zmax-zmin;\
		TIMER_STOP(39);\
       	if(myid==0) fprintf(stdout,"DomainDecomp. CPU= %f \n", ELAPSED_TIME(39));\
	}\
}


#ifdef INCLUDE_TREE_FORCE
    static float efold=0.9;
#else
    static float efold = 0.5;
#endif
void Savingalldata(){
	int suddenstop;
	if(flagsuddenstop(stepcount+1)) suddenstop = 1;
	else suddenstop = 0;
	if( elapsedtime > maxtime|| stepcount%2000 == 0 || suddenstop) {
		delta_a = astep;
		jwrite(stepcount+1);
		if(stepcount%2000!=0) MPI_Finalize();exit(0);
	}
}

void SaveWholeDen(float *fftwmesh,int nx,int ny,int nz,int local_nz,
		int myid,int nid, int stepcount, float amax,float a){
	int i,j,k;
	int ii;
	FILE *fp;
	char filename[100];
	sprintf(filename,"WholeDen.%.5d",stepcount);
	for(ii=0;ii<nid;ii++){
		if(ii == myid){
			if(myid ==0)  fp = fopen(filename,"w");
			else fp = fopen(filename,"a");
			if(myid == 0) {
				fwrite(&amax,sizeof(float),1,fp);
				fwrite(&a,sizeof(float),1,fp);
				fwrite(&omep,sizeof(float),1,fp);
				fwrite(&omegabaryon,sizeof(float),1,fp);
				fwrite(&omeplam,sizeof(float),1,fp);
				fwrite(&hubble,sizeof(float),1,fp);
				fwrite(&bias_8mpc,sizeof(float),1,fp);
				fwrite(&size,sizeof(float),1,fp);
				fwrite(&nx,sizeof(int),1,fp);
				fwrite(&ny,sizeof(int),1,fp);
				fwrite(&nz,sizeof(int),1,fp);
			}
			fwrite(fftwmesh,sizeof(float),2*(nx/2+1)*ny*local_nz,fp);
			fclose(fp);
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}

}

void tree2pmtype(int np){
	int i;
	for(i=0;i<np;i++){
		pmparticles[i].x = treeparticles[i].x;
		pmparticles[i].y = treeparticles[i].y;
		pmparticles[i].z = treeparticles[i].z;
		pmparticles[i].vx = treeparticles[i].vx;
		pmparticles[i].vy = treeparticles[i].vy;
		pmparticles[i].vz = treeparticles[i].vz;
	}
	pmparticles = (pmparticletype*)Realloc(pmparticles,sizeof(pmparticletype)*np);
}
void pm2treetype(int np){
	int i;
	pmparticles = (pmparticletype*)Realloc(pmparticles,sizeof(treeparticletype)*np);
	for(i=np-1;i>=0;i--){
		treeparticles[i].vz = pmparticles[i].vz;
		treeparticles[i].vy = pmparticles[i].vy;
		treeparticles[i].vx = pmparticles[i].vx;
		treeparticles[i].z = pmparticles[i].z;
		treeparticles[i].y = pmparticles[i].y;
		treeparticles[i].x = pmparticles[i].x;
	}
}
void Savexzslice(float *sden,int nx,int ny,int nz,int local_nz, int myid,int nid,
		int stepcount){
	char filename[100];
	FILE *fp;
	float *tden,*allden;
	int i,j,k,ii,mx,slocal_nz;
	int offset;
	MPI_Status status;
	sprintf(filename,"xzslice.%.5d",stepcount);
	mx = 2*(nx/2+1);
	{
		tden = (float*)Malloc(sizeof(float)*nx*local_nz,PPTR(tden));
		for(i=0;i<nx*local_nz;i++) tden[i] = 0.;
		for(k=0;k<local_nz;k++) for(j=0;j<3;j++)for(i=0;i<nx;i++)
			tden[i+nx*k] += sden[i+mx*(j+ny*k)];

	}
	if(myid==0){
		allden=(float*) Malloc(sizeof(float)*nx*nz,PPTR(allden));
		for(i=0;i<nx*local_nz;i++) allden[i] = tden[i];
		offset=local_nz*nx;
		for(i=1;i<nid;i++){
			MPI_Recv(&slocal_nz,1,MPI_INT,i,0,MPI_COMM_WORLD,&status);
			MPI_Recv(allden+offset,
					slocal_nz*nx,MPI_FLOAT,i,1,MPI_COMM_WORLD,&status);
			offset += slocal_nz*nx;
		}
		fp=fopen(filename,"w");
		fwrite(&nx,sizeof(int),1,fp);
		fwrite(&nz,sizeof(int),1,fp);
		fwrite(allden,sizeof(float),nx*nz,fp);
		fclose(fp);
		Free(allden);
	}
	else {
		MPI_Send(&local_nz,1,MPI_INT,0,0,MPI_COMM_WORLD);
		MPI_Send(tden,local_nz*nx,MPI_FLOAT,0,1,MPI_COMM_WORLD);
	}
	/*
	for(ii=0;ii<nid;ii++){
		if(myid == ii){
			if(myid == 0){
				fp = fopen(filename,"w");
				fwrite(&nx,sizeof(int),1,fp);
				fwrite(&nz,sizeof(int),1,fp);
			}
			else fp = fopen(filename,"a");
			fwrite(tden,sizeof(float),nx*local_nz,fp);
			fflush(fp);
			fclose(fp);
		}
		MPI_Barrier(MPI_COMM_WORLD);
	}
	*/
	Free(tden);
}
void Savexyslice(float *sden,int nx,int ny,int nz,int myid,int stepcount){
	char filename[100];
	float *tden;
	FILE *fp;
	int i,j,k,mx;
	int npix;
	if(myid == 0){
		mx = 2*(nx/2+1);
		sprintf(filename,"xyslice.%.5d",stepcount);
		fp = fopen(filename,"w");
		fwrite(&nx,sizeof(int),1,fp);
		fwrite(&ny,sizeof(int),1,fp);
		tden = (float *)Malloc(sizeof(float)*nx*ny,PPTR(tden));
		/*
		for(i=0;i<nx*ny;i++) tden[i] = 0.;
		for(k=0;k<3;k++) for(j=0;j<ny;j++) for(i=0;i<nx;i++)
				tden[i+nx*j] += sden[i+mx*(j+ny*k)];
		fwrite(tden,sizeof(float),nx*ny,fp);
		*/
		for(k=0;k<4;k++) {
			for(j=0;j<ny;j++) for(i=0;i<nx;i++){
				npix = i+mx*(j+ny*k);
				tden[i+nx*j] = sden[npix];
			}
			fwrite(tden,sizeof(float),nx*ny,fp);
		}
		fflush(fp);
		fclose(fp);
		Free(tden);
	}
}
/*
void main(int argc, char *argv[]){
*/
void MAIN_(int argc, char *argv[]){
	int i,j,k;
	int nx, ny, nz, local_nz;
	int nspace;
	int ierror=1, iwrite=0, icont;
    char c;
	int memflag, InitializeBigMemory();
	FILE *simfile;
	rfftwnd_mpi_plan plan,iplan;

	MPI_Init(&argc,&argv);
	MPI_Comm_rank(MPI_COMM_WORLD,&myid);
	MPI_Comm_size(MPI_COMM_WORLD,&nid);

	if( argc == 3 ) {
		if( !(simfile = fopen(argv[1],"r")) ) {
		    fprintf(stderr,"Can't open %s - exiting\n",argv[1]);
			MPI_Finalize();
			exit(0);
		}
		maxtime = atof(argv[2]);
   	}
	else {
		fprintf(stderr,"usage: pyul paramsfile maximum_run_time\n");
		MPI_Finalize();
		exit(0);
	}
	{
		/* reading parameters */
		ReadSimulationParameters(simfile,&icont);
		nx = nxmesh; ny = nymesh; nz = nzmesh;
		nspace = nspace0;
		rth = rtophat_mesh;
		omep = omega;
		omeplam = lambda;
		kenold = ken;
		size = boxsize_hinv_mpc;
		iseed = seed;
		hubble = hubble_parameter;
		amax = 1.0 + z_init;
		astep = delta_a;
		zi = amax - 1.;
	    omei = omep*pow3(1.+zi)/
        	(omep*pow3((1.+zi))+omeplam+(1.-omep-omeplam)*pow2((1.+zi)));
	    redshift = amax/a - 1.0;
	}


	if(Make_Total_Memory() == 0 ){
		fprintf(stderr,"Can't initialize memory - aborting job\n");
		MPI_Finalize();
		exit(0);
	}
	{
		/* determining local_nz for fftw */
		fftwinit_(&nx,&ny,&nz,&local_nz,&local_z_start);
		
	}

	fprintf(stderr,"icont %d\n",icont);
	iwrite = 0;
	if( icont == 0 ) { iwrite = 1; }

	nxpos = (int)offsetof(pmparticletype,x) /sizeof(float);
	nsizepmparticletype = (int)sizeof(pmparticletype) /sizeof(float);

	zheight = (float) local_nz;
	zstart = (float) local_z_start;
	fda4sup_(&nx,&ny,&nz, &local_nz);
	/* PMSEED */
	if(icont ==0){
		float *fx,*fy,*fz,*denpmseedpad;


#ifdef PMSEEDFORCE
		{
			int densize;
			np = (nx/nspace)*(ny/nspace)*(local_nz/nspace);
			pmparticles=(pmparticletype*) Malloc(sizeof(pmparticletype)*np,
					PPTR(pmparticles));
			densize = (2*(nx/2+1))*ny*(local_nz+7);
			den=(float*) Malloc(sizeof(float)*densize, PPTR(den));
			fx = (float *) Malloc(densize*sizeof(float),PPTR(fx));
			fy = (float *) Malloc(densize*sizeof(float),PPTR(fy));
			fz = (float *) Malloc(densize*sizeof(float),PPTR(fz));
			denpmseedpad = (float *) Malloc(2*ny*nz*sizeof(float),
					PPTR(denpmseedpad));
			pmseedforce_(&np,pmparticles,&nsizepmparticletype,&nxpos,den,
					&omep,&omeplam,&kenold,fx,fy,fz,denpmseedpad);
			Free(denpmseedpad);
			Free(fz);Free(fy);Free(fx);
			Free(den);
		}
#else
		{
			int densize;
			int izstart,izwidth,nadd,nzparticles,nzprefixel;
			float *pvmesh;
			float zfinal,vamp,zstartpos,omei;
			zheight = (float)nz/(float)nid; /*equal slab width */
			zstart = zheight*myid;
			zfinal = zheight*(myid+1);
			zheight = zfinal-zstart;

			izstart = (int)zstart;
			izwidth = ceil(zstart+zheight)-izstart;

			nzparticles = 0;
			zstartpos = 2*nz;
			for(i=0;i<nz;i+=nspace){
				if((float)i>=zstart && (float)i<(zstart+zheight)) {
					nzparticles ++;
					zstartpos=min(zstartpos,(float)i);
				}
			}
			nzprefixel = zstartpos-izstart;

			printf("P%d: %g %g nzzzz: %d %d %g\n",myid,zstart,zstart+zheight,
					nzparticles,nzprefixel,zstartpos);

			densize = 2*(nx/2+1)*ny*(local_nz+0);
			if(densize==0) densize=10;
			den=(float*) Malloc(sizeof(float)*densize, PPTR(den));
			denpmseedpad = (float *) Malloc(2*ny*nz*sizeof(float),PPTR(denpmseedpad));
			pmseed_(&np,pmparticles,&nsizepmparticletype,&nxpos,den,
					&omep,&omeplam,&kenold,denpmseedpad,&zstart,&zheight,&vamp,&omei);
			Free(denpmseedpad);
			pvmesh = (float *)
				Malloc(sizeof(float)*2*(nx/2+1)*ny*(ceil(zstart+zheight)-(int)zstart+4),
					PPTR(pvmesh));
			zfinal = zstart+zheight;
			adVarPM_fda4_mesh(den,local_nz,local_z_start,pvmesh,zstart,zfinal,nx,ny,nz);
			Free(den);
			paddingpotentialmesh(pvmesh,izstart,izwidth,izwidth,nx,ny,nz,-2,2);
			np = freespace()/sizeof(pmparticletype) - 32L;
			pmparticles=(pmparticletype*) Malloc(sizeof(pmparticletype)*np,
					PPTR(pmparticles));
			setboundaryparticles_(&np,pmparticles,&nsizepmparticletype,&nxpos,
					&nzparticles,&zstartpos,&nzprefixel,&omep,&omeplam,
					&ken,pvmesh,&zstart,&zheight,&vamp,&nspace,&amax,&omei);
			pmparticles=(pmparticletype*) Realloc(pmparticles,
						sizeof(pmparticletype)*np);
			printf("P%d %g %g %g\n",myid,pmparticles[0].x,pmparticles[0].y,pmparticles[0].z);
			{ size_t sizetnp; sizetnp=np;
				migrate(&sizetnp,nz,zheight,zstart);
				np = sizetnp;
			}
			{
				size_t nmax;
				nmax = freespace()/sizeof(pmparticletype) - 32L;
				pmparticles=(pmparticletype*) Realloc(pmparticles,
						sizeof(pmparticletype)*(np+nmax));
			}
			setinnerparticles_(&np,pmparticles,&nsizepmparticletype,&nxpos,
					&nzparticles,&zstartpos,&nzprefixel,&omep,&omeplam,
					&ken,pvmesh,&zstart,&zheight,&vamp,&nspace,&amax,&omei);
			pmparticles=(pmparticletype*)Realloc(pmparticles,
					sizeof(pmparticletype)*np);
			Free(pvmesh);
		}
#endif
		/*
		lzfloor = local_z_start; lzceil = local_z_start + local_nz;
		*/
		now = 0;
		nowflag = 0;
		{ size_t sizetnp; sizetnp=np;
			migrate(&sizetnp,nz,zheight,zstart);
			np = sizetnp;
		}
        if( myid == 0 ) {
			pmparticletype *bp;
			bp = pmparticles;
			fprintf(stderr,"after migrate\n");
			fprintf(stderr,"P%d particle0 is %g %g %g %g %g %g\n", myid,
				bp->x, bp->y, bp->z, bp->vx, bp->vy, bp->vz);
		}
		{
			long lnp,tlnp;
			lnp = (long ) np;
			MPI_Reduce(&lnp,&tlnp,1,MPI_LONG,MPI_SUM,0,MPI_COMM_WORLD);
			if(myid == 0) fprintf(stderr,"Total particle number : %ld\n",tlnp);
		}
	}
	else if(icont ==1){
		int nmax;
		nmax = freespace()/sizeof(pmparticletype)-1000L;
		pmparticles = (pmparticletype *)Malloc(sizeof(pmparticletype)*nmax,
				PPTR(pmparticles));
		jread();
		pmparticles = (pmparticletype *)Realloc(pmparticles,np*sizeof(pmparticletype));
		{ size_t sizetnp; sizetnp=np;
			migrate(&sizetnp,nz,zheight,zstart);
			np = sizetnp;
		}
	}
	printf("P%d has %d %d %d\n",myid,stepcount,stepnum,nstep);
	{
		long tlnp,lnp;
		float rngc;
		rngc = (float)nx*(float)ny*(float)nz;
		lnp = (long ) np;
		MPI_Reduce(&lnp,&tlnp,1,MPI_LONG,MPI_SUM,0,MPI_COMM_WORLD);
		MPI_Bcast(&tlnp,1,MPI_LONG,0,MPI_COMM_WORLD);
		pmas = rngc/(float)tlnp;

	}
			exit(999);
	/*
	 *  Cell positions run from 0 to n-1 while particle positions
	 *  run from 0 to n
	 *  (xp,yp,zp) = particle position
	 *  CPU : Scalar (1.84), Vector (1.02,8.1,65) for ng = 32,64,128
	 *  */
/* MAIN TIME LOOP FOR SIMULATION */
	/* nzheight : variable width == local_nz_varpm
	 * local_nz : fftw width */
	for(stepcount=stepnum; stepcount<=nstep; stepcount++) {
		int nzheight;
		TIMER_START(51);
		lzfloor = zstart; lzceil = zstart+zheight;

		nzheight = ceil(lzceil) - (int)lzfloor;
		printf("P%d has np=%d %d from %g to %g and local_z_start=%d local_nz=%d\n",myid,np,
				nzheight,lzfloor,lzceil,local_z_start,local_nz);
		{
			float maxd;
			void ExtractingLightConeParticles(pmparticletype *, int ,int ,int ,int , 
					int , float , float , float ,float , float , float ,float ,
					float ,int ,int ,float );
			nextastep = astep;
			maxd = 0.1*nspace;
			ExtractingLightConeParticles(pmparticles, np,nx,ny,nz, stepcount, 
						amax, a, astep,nextastep, size, omep,omeplam,hubble,
						myid,nid,maxd);

		}


		if(flagsyncwholeden(amax,a,astep)){
			halfstep.first = 1;
			halfstep.second = 0;
		}
		else {
			halfstep.first = halfstep.second=0;
		}
halfevolution:
		den = (float *) Malloc(((nzheight+4)*(2*(nx/2+1))*ny)*
				sizeof(float), PPTR(den));
		if(local_nz > 0) fftwmesh = (float *) Malloc(((local_nz+0)*(2*(nx/2+1))*ny)*
				sizeof(float),PPTR(fftwmesh));
		else fftwmesh = (float *) Malloc(sizeof(float),PPTR(fftwmesh));
		VarPM_tsc2(den,nx,ny,nz, lzfloor,lzceil, pmparticles,
				nsizepmparticletype,nxpos, np,pmas,
                fftwmesh, local_nz,local_z_start);
		den = (float*)Realloc(den,(nzheight+3)*(2*(nx/2+1))*ny*sizeof(float));
#ifdef SAVESLICE
		Savexyslice(fftwmesh,nx,ny,nz,myid,stepcount);
		Savexzslice(fftwmesh,nx,ny,nz,local_nz,myid,nid,stepcount);
#endif
		if(flagsyncwholeden(amax,a,astep))
			SaveWholeDen(fftwmesh,nx,ny,nz,local_nz,myid,nid, stepcount,amax,a);
		fftforward_(fftwmesh);
		if(flagpsmeasure(amax,a,astep,stepcount)){
			FILE *fp;
			correl_(fftwmesh,&a,&nowflag,&now,&rth,&nstep);
		}
		psolver_(fftwmesh,&poten,&efold);
		fftbackward_(fftwmesh);
		{
			int nzfloor;
			lzfloor = zstart; lzceil = zstart+zheight;
			adVarPM_fda4_mesh(fftwmesh,local_nz,local_z_start,den,zstart,
					lzceil, nx,ny,nz);
			Free(fftwmesh);
			nzfloor = (int) zstart;
			local_nz_varpm = nzheight;
			den = (float*)Realloc(den,sizeof(float)*2*(nx/2+1)*ny*(nzheight+7));
			/* This is for the padding of boundary slices */
			paddingpotentialmesh(den,nzfloor,nzheight,local_nz_varpm,nx,ny,nz,
					-3, 4);
			MPI_Barrier(MPI_COMM_WORLD);
		}
        fda4sup_(&nx,&ny,&nz,&local_nz_varpm);
		/* WARNING local_nz --> local_nz_varpm*/

		nlzfloor = lzfloor;
		fda4_(pmparticles,&nsizepmparticletype,&nxpos,&np,den,
				&local_nz_varpm,&a,&astep,&nextastep,
				&amax,&omei,&omep,&omeplam,&pfact,&nlzfloor,
				&fact1,&fact2,&(halfstep.first),&(halfstep.second));
		Free(den);
		/* TREE FORCE CORRECTION */
		TIMER_START(59);
		{
			long nplong;
			int treecorrection(treeparticletype *, long,int,int,int,
					float,float,float,float,float,float);
			nplong = np;
			if(stepcount < 150) theta = 0.25;
			else theta = 0.4;
			treeparticles = (treeparticletype *) pmparticles;
			pm2treetype(np);
			(void) treecorrection(treeparticles,nplong, nx, ny, nz,
			  zstart,zheight, fact1, fact2, pfact,theta);
			tree2pmtype(np);
		}
		TIMER_STOP(59);
		if(myid==0) fprintf(stdout,"Treecorrectingtime CPU= %f \n", ELAPSED_TIME(59));
		if(halfstep.first==1 && halfstep.second==0) {
		/* Now positions and velocities of all simulation particles are 
		 * synchronized */
			char saveprefix[80],synchronizervprefix[80];
			strcpy(saveprefix,rvprefix);
			strcpy(synchronizervprefix,rvprefix);
			sprintf(rvprefix,"Sync%s",synchronizervprefix);

			/*
			HyperSort(&np,particles);
			*/
			delta_a = astep;
			jwrite(stepcount);
			strcpy(rvprefix,saveprefix);
		}
		if(halfstep.first==1 && halfstep.second==0) {
			halfstep.second = 1;
			goto halfevolution;
		}
		{
			void SavingLightConeData(pmparticletype *,int ,int ,int , int , int , 
					float , float , float ,float , float ,float , float ,float ,
					float, float ,int ,int);
			SavingLightConeData(pmparticles,np,nx,ny, nz, stepcount, amax, 
					a, astep,nextastep, size, omei, omep,omegabaryon,
					omeplam,hubble,myid,nid);
		}

		/* One step forward with particle positions */
		{
			void onestepforwardposition_(pmparticletype*,int*,int*,int*,float *,
					int *, int *,int *, int *);
			onestepforwardposition_(pmparticles,&np,&nsizepmparticletype,
					&nxpos,&pfact,&nx,&ny,&nz,&stepcount);
		}
		TIMER_STOP(51);
		if(myid==0){
			pmparticletype *bp;
			bp = pmparticles;
			printf("+P%d %g %g %g %g %g %g\n",myid,bp->x,bp->y,bp->z,
					bp->vx,bp->vy,bp->vz);
		}
		{ size_t sizetnp; sizetnp=np;
			migrate(&sizetnp,nz,zheight,zstart);
			np = sizetnp;
		}

		if(myid==0) fprintf(stdout,"Step %d CPU= %f \n", stepcount,ELAPSED_TIME(51));
		if(myid==0){
			elapsedtime = gettime();
			fprintf(stderr,"Wallclock time %d = %g\n",stepcount,elapsedtime);
		}
		MPI_Bcast(&elapsedtime, 1, MPI_FLOAT, 0, MPI_COMM_WORLD);
		a += astep;
		Savingalldata();

		DomainDecomp;

	}



	MPI_Barrier(MPI_COMM_WORLD);
	MPI_Finalize();
	exit(0);
}

int ReadSimulationParameters(simfile,icont)
FILE *simfile;
int *icont;
{
        char filename[180];
	FILE *rvfile;
	int abortflag;

    /* Box size */
	fscanf(simfile,"%f %f\n",&boxsize_hinv_mpc,&hubble_parameter);

	/* Power Spectrum */
	fscanf(simfile,"%f %f %f %f %f %f\n",
           &npow0,&omega,&omegabaryon, &lambda,&bias_8mpc,&rsmooth);
 
	/* Mesh Parameters */
	fscanf(simfile,"%d %d %d %d\n",&nxmesh, &nymesh, &nzmesh,&nspace0);

    /* Tree parameters */
	fscanf(simfile,"%f %f %f\n",&SPHERE_RADIUS,&particle_radius,&theta);
	L = nxmesh; /* Set L parameter */
	L2 = L*L;
		 
	/* Time Parameters */
	fscanf(simfile,"%f %f %f\n",&z_init,&delta_a,&a);
	amax = 1.0 + z_init;

	/* Stepping Parameters */
	fscanf(simfile,"%d %d %d\n",&nstep,&stepnum,&nskip);
		 
	/* Negative integer for random number generator for initial conditions */
    fscanf(simfile,"%d\n",&seed);
		 
	rtophat_mesh = nxmesh/boxsize_hinv_mpc*8.0;

    /* RV file name*/
	fscanf(simfile,"%s %s\n",rvfilename, rvprefix);

    /* Energy parameters */
	fscanf(simfile,"%f %f %f %f\n",
		&ken, &ktot, &const0, &poten0);

    /* Camera file name */
	fscanf(simfile,"%s\n",camerafilename);

    fclose(simfile);

/* Ok now output to stdout to verify all is well */
	if( myid ==0 ) {
		fprintf(stdout,"*** INPUT PARAMETER LIST ***\n\n");
		fprintf(stdout,"*** Cosmological Parameters ***\n");
		fprintf(stdout,"Box size = %g hinv_Mpc   Hubble_parameter = %g\n",
			boxsize_hinv_mpc, hubble_parameter);
		fprintf(stdout,"npow = %g Omega = %g OmegaBaryon = %g Lambda = %g\nBias_8Mpc = %g Rsmooth = %g\n\n", npow0,omega,omegabaryon,lambda,bias_8mpc,rsmooth);
		
		fprintf(stdout,"\n*** Mesh Parameters ***\n");
		fprintf(stdout,"Nxmesh = %d Nymesh = %d Nzmesh = %d Spacing = %d\n\n",
				nxmesh,nymesh,nzmesh,nspace0);
		fprintf(stdout,"\n*** Tree Parameters ***\n");
		fprintf(stdout,"SPHERE_RADIUS = %g Theta = %g\n\n",SPHERE_RADIUS,theta);
		fprintf(stdout,"\n*** Time Parameters ***\n");
		fprintf(stdout,"Initial Redshift = %g Delta_a = %g Current a = %g\n\n",z_init,delta_a,a);
		fprintf(stdout,"\n*** Stepping Parameters ***\n");
		fprintf(stdout,"Nstep = %d Stepnum = %d Nskip = %d\n", 
			nstep,stepnum,nskip);
		fprintf(stdout,"Random Seed = %d\n\n",seed);
		fprintf(stdout,"Initial RV File %s  RV File Prefix %s\n\n",
			rvfilename,rvprefix);
	}
	fflush(stdout);

/* Check to see if the initial conditions file exists - if not then default to g   generating your own initial conditions */

	abortflag = 0;
	if( myid == 0 ) {

/* if strcmp() == 0 then string comparison is successful */
		if( strcmp(rvfilename,"INITIAL") == 0 ) {
			fprintf(stderr,"Generating Initial Conditions for run....\n");
			*icont = 0;
			a = 1.0;
		}
		else {
                        sprintf(filename,"%s%.5d",rvfilename,0);
			rvfile = fopen(filename,"r");
			if( rvfile == NULL ) {
				fprintf(stderr,"Can't open initial rvfile %s - exiting\n",	
					rvfilename);
				abortflag = 1;
			}	
			else { 
				fprintf(stderr,"Successfully opened rvfile %s - starting simulation\n",rvfilename);
				fclose(rvfile); 
				*icont = 1;
			}
		}
	}
	MPI_Bcast(&abortflag, 1, MPI_INT, 0, MPI_COMM_WORLD);
	if( abortflag ) {
		fprintf(stderr,"Aborting can't find initial rvfile\n");
		MPI_Finalize();
		exit(0);
	}

	MPI_Bcast(icont, 1, MPI_INT, 0, MPI_COMM_WORLD);
/* SAMPLE PARAMETER FILE 
 * 65 0.65
 * 1.0 0.3 0.7 1.0 0.0
 * 128 1
 * 4 1.0
 * 71 .10
 * 700 0 100
 * -1
 */
		 
}
			 
/* Initialization for power spectrum and initial condition generator */
			 
void getpowerparams_(nspace,size,amax,astep,npow,smooth,h,rth,bias,iseed,
		omepb)
int *nspace;
float *size, *amax, *astep;
float *npow;
int *iseed;
float *smooth, *h, *rth, *bias;
float *omepb;
{
 
     *nspace = nspace0;
     *size = boxsize_hinv_mpc;
     *amax = 1.0 + z_init;
     *astep = delta_a;
     *npow = npow0;
     *iseed = seed;
     *smooth = rsmooth;
     *h = hubble_parameter;
     *rth = rtophat_mesh;
     *bias = bias_8mpc;
	 *omepb = omegabaryon;
}
