#include "defs.h"

int kchild;
coordStruct c0;

                      /* 3D version:
                             ^ y+                   ^ y+
                             |                      |
                           0 | 1                  4 | 5
                           --+------> x+          ---------> x+
                           2 | 3                  6 | 7
                           
                             +z                     -z
                      */
     /* return the index to specify in which child should the particle go */
#define whichChild(p, c) \
{ \
  kchild = 0; \
  if ((p->r).x - c.x > 0) kchild += 1; \
  if ((p->r).y - c.y > 0) kchild += 2; \
  if ((p->r).z - c.z > 0) kchild += 4; \
}     

     /* coordStruct c; center */
     /* int k;         k-th child */
     /* float d;       half width of cell */
     /* return the coord of the center of the k-th child 
      * of the cell whose center is `c', and for level `l'
      */

#define subCenter(c, k, d) \
{ \
\
  float d0; \
  c0 = c; \
  d0 = d*0.5; \
  if( k & 0x1 ) c0.x = c0.x + d0; \
  else          c0.x = c0.x - d0;\
  if( k & 0x2 ) c0.y = c0.y + d0;\
  else          c0.y = c0.y - d0;\
  if( k & 0x4 ) c0.z = c0.z + d0;\
  else          c0.z = c0.z - d0;\
\
}

nodeStruct  *AddParticleToTree(theNode, theParticle,  center, size)
     nodeStruct *theNode;
     particleStruct *theParticle;
     coordStruct center;
     float size;    /* maximum extent of the cell */
     /* add a particle to the tree below the cell (recursive) */
{
  int i, k, k0, k1;
  int childnum, cellnum, nc;
  coordStruct c;
  nodeStruct *theNewNode, *next, *cfirst, *clast, *child;
  nodeStruct *children[NCHILDREN+1];
  cellStruct *theNewCell;
  cellStruct *MakeNewCell();
  particleStruct *oldParticle;
  particleStruct *p0, *p1;
  float rmin;

  if( theNode == NULL ) {   /* if the node is empty */
    theNewNode = (nodeStruct *) theParticle;  /* insert the particle */ 
    return (theNewNode);
  }

  switch (theNode -> type) {
  case PARTICULE_TYPE:
    /* the current node has only one particle so make a new cell insert
     * the old one in this cell and then insert the new one
     */

	p0 = (particleStruct *) theNode;
    p1 = theParticle;

#ifdef TREEFIX
    rmin = 0.02*sqrt(p0->epssq); /* arbitrary 1/100 the softening length */
	if( rmin > 0.5*size ) { rmin = 0.4*size; }
    if( fabs(p1->r.x - p0->r.x) < rmin )  {
    if( fabs(p1->r.y - p0->r.y) < rmin )  {
    if( fabs(p1->r.z - p0->r.z) < rmin )  {

/* Ok they're very close together - change the input particle position
       slightly i.e. sqrt(3.)/100.*(softening length) */

        fprintf(stderr,"WARNING - perturbing particle with position %g %g %g by %g\n",p1->r.x,p1->r.y,p1->r.z,rmin);
        p1->r.x -= rmin;
		if( p1->r.x < -0.5 )
			p1->r.x += (2.0*rmin);
    }
    }
    }
#endif

	theNewCell = MakeNewCell();   /* make one new cell */
    if( theNewCell == NULL ) {
		return NULL;
	}

    oldParticle = (particleStruct *) theNode; /* save the old particle */
    whichChild(oldParticle, center);  /* find the position on cell */
    k0 = kchild;
    whichChild(theParticle, center);
    k1 = kchild;

/* Insert the new particle in this new cell */

	if( k0 != k1 ) {
		oldParticle->sibling = (nodeStruct *) theParticle;
		theParticle->sibling = NULL;
		oldParticle->cellnum = k0;
		theParticle->cellnum = k1;
		theNewCell->next = (nodeStruct *) oldParticle;
	}
	else { /* k0 == k1  */
		subCenter(center,k0,size); /* calculate new subCell center */
		c = c0;
		next = AddParticleToTree(oldParticle, theParticle, c, 0.5*size);
		next->cellnum = k0;
		theNewCell->next = next;
	}
                                   /* add the particle to this cell */
	theNewNode = (nodeStruct *) theNewCell;
	return theNewNode; /* return the new cell to replace the particle */

  case CELL_TYPE:
/* If the node is a cell then insert the particle */
    whichChild(theParticle, center);
    k = kchild;
	subCenter(center, k, size);
	c = c0;

/* Find the children */
	child = theNode->next;
	nc = 0;
	while( child != NULL ) {
		children[nc] = child;
		child = child->sibling;
		nc++;
	}
	children[nc] = NULL;

/* Scan for a child with the same cell number */
	cfirst = children[0];
	clast  = children[nc-1];
	childnum = -1;
	for(i=0; i<nc; i++) {       /* look for current cell number */
		cellnum = children[i]->cellnum;
		if( cellnum == k ) {
			childnum = i;
			break;
		}
	}

	if( childnum == 0 ) {
	    children[0] =
			AddParticleToTree(children[0], theParticle, c, 0.5*size);
		theNode->next = children[0];
		children[0]->sibling = children[1];
		children[0]->cellnum = k;
	}
	else if ( childnum > 0 ) {
		children[childnum] =
			AddParticleToTree(children[childnum], theParticle, c, 0.5*size);
		children[childnum-1]->sibling = children[childnum];
		children[childnum]->sibling = children[childnum+1];
		children[childnum]->cellnum = k;
	}
	else {  /* childnum = -1 - adding a particle to a NULL child */
		child = AddParticleToTree((nodeStruct *) 0, theParticle, c, 0.5*size);
		children[nc-1]->sibling = child;
		child->sibling = NULL;
		child->cellnum = k;
	}
		
    return(theNode); /* return the old node */


  default:
    fprintf(stderr,"AddParticleToTree: found an invalid cell type\n");
    EXIT(11);
  }
}
 
cellStruct *BuildTree(ParticleBin, theta, domainCorner, length)
     /* build a tree for the given 
      * return the pointer to the root-cell of that tree */
particleStruct *ParticleBin;
float theta;
coordStruct domainCorner;
float length;
{
  int i;
  nodeStruct *theRoot;
  nodeStruct *LinkNextNode();
  nodeStruct *PruneTree();
  cellStruct *theRootCell;
  particleStruct *p;
  coordStruct c;
  coordStruct rmin, rmax, rcm, drc;
  coordStruct domainLength;
  int trouble;
  float cellsq, delta;
  float systemSize;  /* Maximum extent of the system */

	domainLength.x = length;
	domainLength.y = length;
	domainLength.z = length;
  systemSize = domainLength.x;
  if( systemSize < domainLength.y )
    systemSize = domainLength.y;
  if( systemSize < domainLength.z )
    systemSize = domainLength.z;
  cellsq = systemSize*systemSize;


/*
 * always start at the corner of the root cell
 */

  c.x = domainCorner.x;
  c.y = domainCorner.y;
  c.z = domainCorner.z;

#ifdef TESTING
  BinExtent(ParticleBin,&rmin,&rmax);
  trouble = 0;
  if( c.x > rmin.x ) trouble=1;
  if( c.y > rmin.y ) trouble=1;
  if( c.z > rmin.z ) trouble=1;
  if( c.x + systemSize < rmax.x ) trouble=1;
  if( c.y + systemSize < rmax.y ) trouble=1;
  if( c.z + systemSize < rmax.z ) trouble=1;
  if( trouble ) {
     fprintf(stderr,"corner %g %g %g size %g\n",c.x,c.y,c.z,systemSize);
     fprintf(stderr,"max %g %g %g\n",
			c.x+systemSize,c.y+systemSize,c.z+systemSize);
     fprintf(stderr,"rmin %g %g %g\n",rmin.x,rmin.y,rmin.z);
     fprintf(stderr,"rmax %g %g %g\n",rmax.x,rmax.y,rmax.z);
EXIT(0);
  }
#endif

  theRoot = NULL; /* deallocate the current tree */

  /* insert the particles one at a time */
	i = 0;
	for(p=ParticleBin; p != NULL; p=p->nextinbin) {
		if( p->mass == 0.0 ) continue;
		theRoot = AddParticleToTree(theRoot, p, c, systemSize);
		if( theRoot == NULL )
			return NULL;
		i++;
	}

  theRootCell = (cellStruct *) theRoot;
/*
	WalkTree(theRoot,0);
*/
  CountChildren(theRoot);
  theRootCell->sibling = NULL;  /* the root has no siblings */
  LinkSiblings(theRootCell);    /* link all the siblings */

  FindCenters(theRoot);     /* Find the center of mass of the cells */
  FindQuadrupoles(theRoot); /* Find quadrupole moments of cells */

  theRootCell->corner = domainCorner;  /* reassign corner */
  theRootCell->length = domainLength;  /* reassign lengths to root cell */
  theRootCell->cellsq = systemSize*systemSize;
  theRootCell->radius = sqrt(systemSize*systemSize);
#ifdef NEW_BARNES_HUT
  rcm = theRootCell->r;
  drc.x = c.x - rcm.x;
  drc.y = c.y - rcm.y;
  drc.z = c.z - rcm.z;
  delta = sqrt(drc.x*drc.x + drc.y*drc.y + drc.z*drc.z);
  cellsq = systemSize/theta + delta;
  theRootCell->cellsq = cellsq*cellsq;
#endif

	CellExtent(theRootCell,theta,&rmin,&rmax);

/*  FindDimensions(theRootCell,theta,0.5*systemSize,c); */
/*  LookTree(theRoot); */
/*  theRoot = PruneTree(theRoot); */

  return theRootCell;
}

nodeStruct *FindLastCell(theRoot)
cellStruct *theRoot;
{
	nodeStruct *nextNode, *lastNode;
	particleStruct *p;
	cellStruct *c;

	lastNode = (nodeStruct *) theRoot;
	nextNode = (nodeStruct *) theRoot;
	while( nextNode != NULL ) {
		if( nextNode->type == PARTICULE_TYPE ) {
			p = (particleStruct *) nextNode;
			lastNode = nextNode;
			nextNode = p->sibling;
		}
		else if( nextNode->type == CELL_TYPE ) {
			c = (cellStruct *) nextNode;
			lastNode = nextNode;
			nextNode = c->next;
		}
	}
	return lastNode;
}

FindCenters(theNode)
nodeStruct *theNode;
{
	int i, ncount, nchildren;
	float xcm, ycm, zcm;
	float meanepssq;
	float mass;
	cellStruct *theCell, *theChild;

/* Assumes a call to CountChildren previously */

	if ( theNode->type == CELL_TYPE ) {
		theCell = (cellStruct *) theNode;
		mass = xcm = ycm = zcm = 0.0;
		meanepssq = 0;
		ncount = 0;
		theChild = (cellStruct *) (theCell->next);
		nchildren = theCell->nchildren;
		for(i=0; i<nchildren; i++) {
			float mc; 

			if( theChild->type == CELL_TYPE )
				FindCenters(theChild);

			mc = theChild->mass;
			xcm += mc*theChild->r.x;
			ycm += mc*theChild->r.y;
			zcm += mc*theChild->r.z;
			mass += mc;
			meanepssq += mc*theChild->epssq;
			if( theChild->type == CELL_TYPE )
				ncount += theChild->nbodies;
			else
				ncount++; /* its a single particle */
			theChild = (cellStruct *) (theChild->sibling);
		}
		xcm /= mass; ycm /= mass; zcm /= mass;
		theCell->mass = mass;
		theCell->r.x = xcm;
		theCell->r.y = ycm;
		theCell->r.z = zcm;
		theCell->nbodies = ncount;
		meanepssq /= mass;
		theCell->epssq = meanepssq;
    }
}

FindQuadrupoles(theNode)
nodeStruct *theNode;
{
	int i, nchildren;
	float mass, mass3;
	float x, y, z, x2, y2, z2;
	cellStruct *theChild;
	cellStruct *theCell;

	if( theNode->type == CELL_TYPE) {
		theCell = (cellStruct *) theNode;

		theCell->qxx = 0.0;
		theCell->qyy = 0.0;
		theCell->qzz = 0.0;
		theCell->qxy = 0.0;
		theCell->qxz = 0.0;
		theCell->qyz = 0.0;
		theCell->mom = 0.0;

		theChild = (cellStruct *) (theCell->next);
		nchildren = theCell->nchildren;
	    for( i = 0 ; i < nchildren ; i++ ) { /* Loop over subcells. */

			if( theChild->type == CELL_TYPE ) {
	            FindQuadrupoles(theChild); /* Determine the quad. tensor */
			
				theCell->qxx += theChild->qxx;
				theCell->qyy += theChild->qyy;
				theCell->qzz += theChild->qzz;
				theCell->qxy += theChild->qxy;
				theCell->qxz += theChild->qxz;
				theCell->qyz += theChild->qyz;
				theCell->mom += theChild->mom;
			}
/* If the child is a particle or a cell you still calculate this */
			x = theChild->r.x - theCell->r.x;
			y = theChild->r.y - theCell->r.y;
			z = theChild->r.z - theCell->r.z;
	
			x2 = x*x;
			y2 = y*y;
			z2 = z*z;

			mass = theChild->mass;
			mass3 = 3.0 * mass;
			theCell->qxx += mass * (2.0*x2 - y2 - z2);
			theCell->qyy += mass * (2.0*y2 - x2 - z2);
			theCell->qzz += mass * (2.0*z2 - x2 - y2);      
			theCell->qxy += mass3 * x*y; 
			theCell->qxz += mass3 * x*z;             
			theCell->qyz += mass3 * y*z;            
			theCell->mom += mass * (x2 + y2 + z2);
			theChild = (cellStruct *) (theChild->sibling);
		}
    }
}

#ifdef VERYOLD
nodeStruct *PruneTree(theNode)
nodeStruct *theNode; /* this node is *always* a CELL */
{
	nodeStruct *theChild, *oldNode;
	cellStruct *oldCell, *theCell, *theSubCell;
    int branch;
    int k, nchild;

	theCell = (cellStruct *) theNode;

    branch = 1;
    nchild = 0;
    for(k=0; k<NCHILDREN; k++) {         /* Loop thru children */
        theChild = theCell->child[k];
        if( theChild != NULL ) {
            if( theChild->type == CELL_TYPE ) {
                oldNode = PruneTree(theChild); /* if its a cell prune it */
                nchild++;             /* count number of CELL children */
            }
            else 
				branch = 0; /* Its a body not a branch */
        }
    }

    if( branch ) {  /* If its a cell in the tree */
        if( nchild == 1 ) { /* and it only has one subcell */
            return(oldNode);
				/* snip off the cell and graft its subcell */
                /* onto its position in the tree */
        }
    }
	return(theNode);
}
#endif

LinkSiblings(theNode)
nodeStruct *theNode;
{
    int i, nchildren;
	nodeStruct *theChild;
	cellStruct *theCell;

	theCell = (cellStruct *) theNode;
	nchildren = theCell->nchildren;
	theChild = theCell->next;
    for(i=0; i<nchildren-1; i++) {  /* Loop thru subnodes */
		theChild = theChild->sibling;
    }
	theChild->sibling = theCell->sibling;

	theChild = theCell->next;
    for(i=0; i<nchildren; i++) {         /* walk down tree to link children */
		if( theChild->type == CELL_TYPE )
			LinkSiblings(theChild);
		theChild = theChild->sibling;
	}
}

#ifdef VERYOLD
nodeStruct *LinkNextNode(theNode)
nodeStruct *theNode;
{
	int i, j;
	nodeStruct *theChild[8];
	cellStruct *theCell, *theSubCell;
	particleStruct *theParticle;

	theCell = (cellStruct *) theNode;

	for(i=0, j=0; i<NCHILDREN; i++) {
		if( theCell->child[i] != NULL )
			theChild[j++] = theCell->child[i];
	}

	for(i=0; i<j; i++) {
		if( theChild[i]->type == CELL_TYPE ) {
			theSubCell = (cellStruct *) theChild[i];
			theSubCell->next = LinkNextNode(theChild[i]);
		}
	}

	return(theChild[0]); /* return the first child */
}

LookTree(theNode)
nodeStruct *theNode;
{
	int i, np, nc;
	cellStruct *theCell, *c[8];
	particleStruct *p[8];

	theCell = (cellStruct *) theNode;

	np = nc = 0;
	for(i=0; i<NCHILDREN; i++) {
		nodeStruct *theChild;

		theChild = theCell->child[i];
		if( theChild == NULL )
			continue;
		if( theChild->type == PARTICULE_TYPE ) {
			p[np] = (particleStruct *) theChild;
			np++;
		}
		else if ( theChild->type == CELL_TYPE ) {
			c[nc] = (cellStruct *) theChild;
			nc++;
		}
	}
	for(i=0; i<nc; i++)
		LookTree(c[i]);
}
#endif

/* Count the number of children that each cell has */

CountChildren(theCell)
cellStruct *theCell;
{
	int nc;
	nodeStruct *theChild;

	nc = 0;
	theChild = theCell->next;
	while( theChild != NULL ) {
		theChild = theChild->sibling;
		nc++;
	}
	theCell->nchildren = nc;

	theChild = theCell->next;
	while( theChild != NULL ) {
		if( theChild->type == CELL_TYPE )
			CountChildren(theChild);
		theChild = theChild->sibling;
	}
}

FindDimensions(theCell,theta,size,center)
cellStruct *theCell;
float theta, size;
coordStruct center;
{
	int i, nchildren, cellnum;
	nodeStruct *theChild;
	cellStruct *theSubCell;
	coordStruct rcellmin, rmax, rcellmax;
	coordStruct corner, length, c, rcm, drc;
	float delta, cellsq;

	nchildren = theCell->nchildren;
	theChild = theCell->next;
	for(i=0; i<nchildren; i++) {
		corner = theCell->corner;
		length = theCell->length;
		rmax.x = corner.x + length.x;
		rmax.y = corner.y + length.y;
		rmax.z = corner.z + length.z;
		rcellmin.x = center.x - size;
		rcellmin.y = center.y - size;
		rcellmin.z = center.z - size;
		rcellmax.x = center.x + size;
		rcellmax.y = center.y + size;
		rcellmax.z = center.z + size;

		if( theChild->type == CELL_TYPE ) {
			theSubCell = (cellStruct *) theChild;
			cellnum = theSubCell->cellnum;
			if( cellnum & 0x1 ) rcellmin.x += size; 
			else          rcellmax.x -= size;
			if( cellnum & 0x2 ) rcellmin.y += size;
			else          rcellmax.y -= size;
			if( cellnum & 0x4 ) rcellmin.z += size;
			else          rcellmax.z -= size;
			if( rcellmin.x > corner.x ) corner.x = rcellmin.x;
			if( rcellmin.y > corner.y ) corner.y = rcellmin.y;
			if( rcellmin.z > corner.z ) corner.z = rcellmin.z;
			if( rcellmax.x < rmax.x ) rmax.x = rcellmax.x;
			if( rcellmax.y < rmax.y ) rmax.y = rcellmax.y;
			if( rcellmax.z < rmax.z ) rmax.z = rcellmax.z;
			length.x = rmax.x - corner.x;
			length.y = rmax.y - corner.y;
			length.z = rmax.z - corner.z;
			c.x = 0.5*(rcellmin.x + rcellmax.x);
			c.y = 0.5*(rcellmin.y + rcellmax.y);
			c.z = 0.5*(rcellmin.z + rcellmax.z);
			theSubCell->corner = corner;
			theSubCell->length = length;
			theSubCell->cellsq = size*size;
#ifdef NEW_BARNES_HUT
			rcm = theSubCell->r;
			drc.x = c.x - rcm.x;
			drc.y = c.y - rcm.y;
			drc.z = c.z - rcm.z;
			delta = sqrt(drc.x*drc.x + drc.y*drc.y + drc.z*drc.z);
			cellsq = size/theta + delta;
			theSubCell->cellsq = cellsq*cellsq;
#endif
			FindDimensions(theSubCell,theta,0.5*size,c);
		}
		theChild = theChild->sibling;
	}
}

CellExtent(theCell,theta,rmin,rmax)
cellStruct *theCell;
float theta;
coordStruct *rmin, *rmax;
{
	int i, nchildren;
	coordStruct r0, r1;
	coordStruct r0c, r1c, length, r, drc, rcm, center;
	float cellsq, size, delta;
	float dx0, dx1, dy0, dy1, dz0, dz1, dxm, dym, dzm;
	nodeStruct *theChild;
	particleStruct *p;

	nchildren = theCell->nchildren;
	theChild = theCell->next;
	r0.x = r0.y = r0.z = 1.0e30;
	r1.x = r1.y = r1.z = -1.0e30;
	for(i=0; i<nchildren; i++) {
		switch( theChild->type ) {
			case PARTICULE_TYPE:
			p = (particleStruct *) theChild;
			r = p->r;
			if( r.x < r0.x ) r0.x = r.x;
			if( r.y < r0.y ) r0.y = r.y;
			if( r.z < r0.z ) r0.z = r.z;
			if( r.x > r1.x ) r1.x = r.x;
			if( r.y > r1.y ) r1.y = r.y;
			if( r.z > r1.z ) r1.z = r.z;
			break;

			case CELL_TYPE:
			CellExtent(theChild,theta,&r0c,&r1c);
			if( r0c.x < r0.x ) r0.x = r0c.x;
			if( r0c.y < r0.y ) r0.y = r0c.y;
			if( r0c.z < r0.z ) r0.z = r0c.z;
			if( r1c.x > r1.x ) r1.x = r1c.x;
			if( r1c.y > r1.y ) r1.y = r1c.y;
			if( r1c.z > r1.z ) r1.z = r1c.z;
			break;
		}
		theChild = theChild->sibling;
	}
	*rmin = r0;
	*rmax = r1;
	theCell->corner = r0;
	length.x = r1.x - r0.x;
	length.y = r1.y - r0.y;
	length.z = r1.z - r0.z;
	theCell->length = length;

	size = length.x;
	if( size < length.y )
		size = length.y;
	if( size < length.z )
		size = length.z;
	theCell->cellsq = size*size;
#ifdef NEW_BARNES_HUT
	rcm = theCell->r;
	center.x = 0.5*(r0.x + r1.x);
	center.y = 0.5*(r0.y + r1.y);
	center.z = 0.5*(r0.z + r1.z);
    drc.x = center.x - rcm.x;
    drc.y = center.y - rcm.y;
    drc.z = center.z - rcm.z;
    delta = sqrt(drc.x*drc.x + drc.y*drc.y + drc.z*drc.z);
    cellsq = size/theta + delta;
    theCell->cellsq = cellsq*cellsq;
#endif

/*
 * Determine a "radius" for the cell with respect to the centre of mass 
 */
	rcm = theCell->r;
	dx0 = rcm.x - r0.x; dx1 = r1.x - rcm.x; 
	dxm = dx0; if( dx1 > dxm ) dxm = dx1;

	dy0 = rcm.y - r0.y; dy1 = r1.y - rcm.y; 
	dym = dy0; if( dy1 > dym ) dym = dy1;

	dz0 = rcm.z - r0.z; dz1 = r1.z - rcm.z; 
	dzm = dz0; if( dz1 > dzm ) dzm = dz1;

	theCell->radius = sqrt(dxm*dxm + dym*dym + dzm*dzm);
	
}
	

#ifdef INTERACT
WalkTree(theNode,level)
int level;
nodeStruct *theNode;
{
	int next=1, childcount, l;
	nodeStruct *s, *n;

	next = 0;
	childcount = 0;
	l = level;
	while( 1 ) {
		fprintf(stdout,"type %d level %d cellnum %d childcount %d\n",
			theNode->type,l,theNode->cellnum, childcount);
		iquery("Next or Sibling or up ?",&next);
		if( next == 0 )  {
			if( theNode->type == PARTICULE_TYPE )
				fprintf(stderr,"particle -can't go down\n");
			else {
				fprintf(stderr,"going down\n");
				WalkTree(theNode->next,l+1);
			}
		}
		else if( next == 1 ) {
			if( theNode->sibling == NULL )
				fprintf(stderr,"last sibling\n");
			else {
				fprintf(stderr,"going to sibling\n");
				WalkTree(theNode->sibling,level);
				childcount++;
			}
		}
		else break; /* go back up */
	}
}
#endif
